#!/usr/bin/python3

# vim: set expandtab ts=4 sw=4;

from csv import DictReader
from os import makedirs, symlink
from os.path import join
from shutil import copyfile
import warnings

import tempfile

import numpy as np

import matplotlib.colors as mc
import matplotlib.pyplot as plt

# Note that we currently assume the use of 1000 for band widths to make life
# easier
BAND_WIDTH = 1000

CIRCOS_CONF_TEMPLATE = """
karyotype = {karyotype:s}
chromosomes_units = 1

<ideogram>
<spacing>
default = 0.005r
</spacing>
radius           = {circle_radius:f}r
thickness        = 50p
fill             = yes
stroke_color     = dgrey
stroke_thickness = 2p
show_label       = no
show_bands            = yes
fill_bands            = yes
band_stroke_thickness = 2
band_stroke_color     = white
band_transparency     = 4
</ideogram>

{colors:s}

{ticks:s}

<image>
angle_orientation* = counterclockwise
radius* = {radius:d}p
<<include {circospath:s}/image.conf>>
</image>
<<include {circospath:s}/colors_fonts_patterns.conf>>
<<include {circospath:s}/housekeeping.conf>>
<links>
<link>
file          = {linksname:s}
radius        = 0.95r
color         = black_a4
bezier_radius = 0.1r
thickness     = 1
</link>
</links>

<image>
file = {outputimage:s}
</image>
"""


class CircosException(BaseException):
    def __init__(self, retcode, stdout, stderr):
        self.retcode = retcode
        self.stdout = stdout
        self.stderr = stderr

    def __str__(self):
        return '<CircosException: Return code: {}>'.format(self.retcode)


class CircosGroup(object):
    def __init__(self, name, c_idx=0, n_idx=0, circoscolour='unknown',
                 rgbcolour=None, netmatreverse=False):

        self.name = name
        self.c_idx = c_idx
        self.n_idx = n_idx
        self.circoscolour = circoscolour
        if rgbcolour is None:
            self.rgbcolour = [0, 0, 0]
        else:
            # Make sure to take a copy
            self.rgbcolour = rgbcolour[:]
        self.netmatreverse = netmatreverse

        # Dictionary of regions
        self.regions = {}

    def __str__(self):
        return '<CircosGroup {}>'.format(self.name)

    def __repr__(self):
        return '<CircosGroup {}>'.format(self.name)

    def __len__(self):
        return len(self.regions)

    def add_region(self, name, offset, shortname, analysisid):
        """
        Adds a new region to the group.

        Parameters
        ----------
        name: str
            Region name

        offset: int
            Region offset value

        shortname: str
            Region short name (used in plotting)

        analysisid: int
            Data index to use when accessing region data
        """
        self.regions[name] = {}
        self.regions[name]['offset'] = offset
        self.regions[name]['shortname'] = shortname
        self.regions[name]['analysisid'] = analysisid

    def find_region_idx(self, region):
        """
        Find a region index from an name.

        Parameters
        ----------
        region: str
            Region name

        Returns
        -------
        int or None
            Region index or None if not found
        """
        if region in self.regions:
            return self.regions[region]['analysisid']
        return None

    def find_region_from_idx(self, idx):
        """
        Find a region name from an index.

        Parameters
        ----------
        idx: int
            Region index

        Returns
        -------
        str or None
            Region name or None if not found
        """
        for region in self.regions.keys():
            if self.regions[region]['analysisid'] == idx:
                return region

        return None

    def has_region(self, region):
        return region in self.regions.keys()

    @property
    def circular_indices(self):
        """
        Get a list of all region analysis indices in circular plotting order.

        Returns
        -------
        list
            List of indices to use when accessing data in order to access
            data in the correct order for circular plotting.
        """
        return [x[1]['analysisid'] for x in sorted(self.regions.items(),
                                                   key=lambda x: x[1]['offset'])]

    @property
    def circular_regions(self):
        """
        Get a list of the regions in circular plotting order

        Returns
        -------
        list
            List of region names to use in the correct order for circular
            plotting.
        """
        return [x[0] for x in sorted(self.regions.items(),
                                     key=lambda x: x[1]['offset'])]

    @property
    def circular_shortnames(self):
        """
        Get a list of the region short names in circular plotting order

        Returns
        -------
        list
            List of region short names names in the correct order for circular
            plotting.
        """
        return [x[1]['shortname'] for x in sorted(self.regions.items(),
                                                  key=lambda x: x[1]['offset'])]

    def circular_link_cache(self, groupname):
        """Return a dictionary of the regions and their information"""
        ret = {}
        for region in self.regions:
            ret[region] = (groupname,
                           self.regions[region]['offset'],
                           self.regions[region]['offset'] + BAND_WIDTH)

        return ret

    @property
    def netmat_indices(self):
        """
        Get a list of all analysis indices in netmat plotting order.

        Returns
        -------
        list
            List of indices to use when accessing data in order to access
            data in the correct order for netmat plotting.
        """

        tmp = self.circular_indices

        if self.netmatreverse:
            tmp = tmp[::-1]

        return tmp

    @property
    def netmat_regions(self):
        """
        Get a list of all region names in netmat plotting order.

        Returns
        -------
        list
            List of names to use when accessing data in order to access
            data in the correct order for netmat plotting.
        """
        tmp = self.circular_regions

        if self.netmatreverse:
            tmp = tmp[::-1]

        return tmp

    @property
    def netmat_shortnames(self):
        """
        Get a list of the region short names in netmat plotting order

        Returns
        -------
        list
            List of region short names names in the correct order for netmat
            plotting.
        """
        tmp = self.circular_shortnames

        if self.netmatreverse:
            tmp = tmp[::-1]

        return tmp


class CircosHandler(object):
    """
    Handler for producing plots using the Circos tool.

    Note that if you are creating plots for publication using Circos, you should
    cite the relevant publication: [Krzywinski2009]_.
    """

    def __init__(self):
        self.groups = {}

    def has_region(self, region):
        """Check whether a region exists in any currently defined group"""
        for region in self.groups.keys():
            if self.groups[region].has_region(region):
                return True

        return False

    def find_region_idx(self, region):
        """Find a region index"""
        for group in self.groups.keys():
            idx = self.groups[group].find_region_idx(region)
            if idx is not None:
                return idx

        return None

    def find_region_from_idx(self, idx):
        """Find a region name from an index"""
        for group in self.groups.keys():
            name = self.groups[group].find_region_from_idx(idx)
            if name is not None:
                return name

        return None

    @property
    def num_regions(self):
        return len(self.circular_regions)

    def colour_mapping(self, normalise=True):
        """
        Return a dictionary of colour name to 3-tuples.

        Parameters
        ----------
        normalise: bool, optional
            If True, colour names will be scaled from the standard 0-255
            range to 0-1; this is primarily useful for matplotlib

        Returns
        -------
        dict
            Maps colour names to 3-tuples of colour information.
        """
        colours = {}
        for group in self.groups.keys():
            colourname = self.groups[group].circoscolour
            rgbcolour = self.groups[group].rgbcolour

            if normalise:
                rgbcolour = [(x / 255.0) for x in rgbcolour]

            colours[colourname] = rgbcolour

        return colours

    @property
    def data_regions(self):
        """Return a list of region names in data order"""
        ret = []
        for k in range(self.num_regions):
            ret.append(self.find_region_from_idx(k))
        return ret

    @property
    def circular_indices(self):
        """Return a list of all region analysis indices in circular plotting order"""
        ret = []
        for group in self.circular_groups:
            ret.extend(self.groups[group].circular_indices)

        return ret

    @property
    def circular_regions(self):
        """Return a list of the groups in ascending order"""
        ret = []
        for group in self.circular_groups:
            ret.extend(self.groups[group].circular_regions)

        return ret

    @property
    def circular_shortnames(self):
        """Return a list of all shortnames in circular plotting order"""
        ret = []
        for group in self.circular_groups:
            ret.extend(self.groups[group].circular_shortnames)

        return ret

    @property
    def circular_link_cache(self):
        """Returns a dictionary which caches the circular position information for links"""
        ret = {}
        for group in self.circular_groups:
            ret.update(self.groups[group].circular_link_cache(group))

        return ret

    @property
    def netmat_indices(self):
        """Return a list of all region analysis indices in netmat plotting order"""
        ret = []
        for group in self.netmat_groups:
            ret.extend(self.groups[group].netmat_indices)

        return ret

    @property
    def netmat_regions(self):
        """Return a list of all regions in netmat plotting order"""
        ret = []
        for group in self.netmat_groups:
            ret.extend(self.groups[group].netmat_regions)

        return ret

    @property
    def netmat_shortnames(self):
        """Return a list of all shortnames in netmat plotting order"""
        ret = []
        for group in self.netmat_groups:
            ret.extend(self.groups[group].netmat_shortnames)

        return ret

    @property
    def circular_groups(self):
        """Return a list of the groups in circular order"""
        return [x[0] for x in sorted(self.groups.items(),
                                     key=lambda x: x[1].c_idx)]

    @property
    def netmat_groups(self):
        """Return a list of the groups in netmat order"""
        return [x[0] for x in sorted(self.groups.items(),
                                     key=lambda x: x[1].n_idx)]

    @classmethod
    def from_csv_files(cls, group_filename, region_filename, analysis_column='AnalysisID'):
        """
        This routine sets up a CircosHandler from two CSV files.

        The group file must contain a list of groups; in Circos terms these are
        known as the 'chromosomes'.  This file must contain the columns:
          * Group: The name of the group: must be unique
          * CircularIdx: The ordering of the group (from 1..N) in the circular plot order

        The following columns are optional:
          * NetmatIdx: The ordering of the group (from 1..N) in the netmat plot order.
                       If not provided, the CircularIdx will be used
          * CircosColour: The colour to use for the chromosome in circos format
          * RGBColour: A 3-tuple of the RGB (0-255) colour to use in non-circos (e.g. for netmats)
                       If no colour is set, mid-grey will be used.
          * NetmatReverseOrder: If True, when sorting out netmat indices,
                                the ordering within the group will be reversed.  This is
                                commonly used when going from circular to netmat order
                                on the left hand side of the plot.  If not set, False
                                will be assumed.

        The region file must contain a list of regions; in Circos
        terms these are known as the 'bands' within each chromosome.
        This file must contain the columns:
          * Region: The name of the region
          * Group: The name of the group which the region should be placed in
          * GroupOffset: The integer location at which the region should be
                         placed within the group.  Note that by default, this
                         should be a multiple of 1000.  The first value within
                         a region should be 0
          * AnalysisID: The position of the data for the region in the analysis
                        data.  I.e. an index from 0-(N-1).  This is used to
                        extract information from, e.g. netmats in the correct
                        order.
                        Note that the name of this column can be varied using
                        the optional 'analysis_column' argument to the function.
                        This allows having CSV files with multiple data orderings
                        in different columns.

        The following columns are optional:
          * ShortName: The short name to use on the outer labels.  If this is
                       not present, the Region name will be used.

        Parameters
        ----------
        group_filename: str
            Filename of group CSV file to load (see above)

        region_filename: str
            Filename of the region CSV file to load (see above)

        analysis_column: str, optional
            Column name to use for loading analysis ordering (see above).
            Defaults to 'AnalysisID'

        Returns
        -------
        CircosHandler
            CircosHandler object ready to be used for plotting.
        """
        ret = cls()

        # Read the groups
        with open(group_filename, 'r') as f:
            reader = DictReader(f)
            for row in reader:
                name = row['Group']

                if name in ret.groups:
                    raise Exception("Group name {} defined more than once".format(name))

                c_idx = int(row['CircularIdx'])
                n_idx = int(row.get('NetmatIdx', c_idx))
                circoscolour = row.get('CircosColour', 'grey')
                rgbcolour = [int(x.strip()) for x in row.get('RGBColour', '128,128,128').split(',')]
                tmp = row.get('NetmatReverseOrder', 'False')
                if tmp.lower() in ['true', '1', 't', 'y', 'yes']:
                    netmatreverse = True
                else:
                    netmatreverse = False

                group = CircosGroup(name, c_idx, n_idx, circoscolour, rgbcolour, netmatreverse)

                ret.groups[name] = group

        # Now read each region
        with open(region_filename, 'r') as f:
            reader = DictReader(f)
            for row in reader:
                name = row['Region']
                group = row['Group']
                offset = int(row['GroupOffset'])
                shortname = row.get('ShortName', name)
                analysispos = int(row.get(analysis_column))

                if group not in ret.groups:
                    raise Exception("Group {} for ROI {} not defined".format(group, name))

                ret.groups[group].add_region(name, offset, shortname, analysispos)

        return ret

    def karyotype(self):
        """Return a string containing the contents of a karyotype file"""
        s = []

        # First of all the group definitions
        for group in self.circular_groups:
            g = self.groups[group]
            num_entries = len(g.regions)
            s.append('chr - {} {} 0 {} {}'.format(group, g.c_idx,
                                                  num_entries * BAND_WIDTH,
                                                  g.circoscolour))

        # Now the region definitions
        for group in self.circular_groups:
            g = self.groups[group]
            for region in g.circular_regions:
                start = g.regions[region]['offset']
                end = start + BAND_WIDTH

                s.append('band {} {} {} {} {} grey'.format(group, region, region,
                                                           start, end))

        return '\n'.join(s)

    def ticks(self, **kwargs):
        """
        Return a string containing the contents of a ticks.conf file
        All parameters must be passed as kwargs.

        Parameters
        ----------
        ticks: bool
            Show ticks around circle

        tick_labels: bool
            Show tick labels around circle

        Returns
        -------
        str
            String containing ticks.conf content needed for the relevant
            plot configuration
        """
        s = []

        # A header
        s.append('show_ticks          = {}'.format('yes' if kwargs.get('ticks', True) else 'no'))
        s.append('show_tick_labels    = {}'.format('yes' if kwargs.get('tick_labels', True) else 'no'))
        s.append('')
        s.append('<ticks>')
        s.append('chromosomes_display_default = no')
        s.append('skip_first_label = no')
        s.append('skip_last_label  = no')
        s.append('radius           = dims(ideogram,radius_outer)')
        s.append('tick_separation  = 2p')
        s.append('label_separation = 5p')
        s.append('multiplier       = 1')
        s.append('color            = black')
        s.append('thickness        = 4p')
        s.append('size             = 20p')
        s.append('show_label     = yes')
        s.append('label_size     = 40p')
        s.append('label_offset   = 10p')

        # Now the region definitions
        for group in self.circular_groups:
            g = self.groups[group]
            for region in g.circular_regions:
                shortname = g.regions[region]['shortname']
                start = int(g.regions[region]['offset'])
                s.append('<tick>')
                s.append('chromosomes = {}'.format(group))
                s.append('position = {}u'.format(start + (BAND_WIDTH // 2)))
                s.append('format = {}'.format(shortname))
                s.append('</tick>')

        s.append('</ticks>')

        return '\n'.join(s)

    def links(self, data, linkcolors=None, defaultlinkcolor='black'):
        """
        Parameters
        ----------
        data: ndarray
            an (N x N) matrix containing thickness (and np.nan where no connection is desired)

        linkcolors: tuple, None or numpy array, optional
            One of 1) a tuple of three strings (zerocolor, negcolor, poscolor);
            e.g. ('black', 'blue', 'red), 2) None (in which case ('black', 'blue', 'red')
            will be assumed) or 3) a numpy array of dtype object (N, N)
            with a specific color string in each entry

            In the tuple case, all data values which are identically 0 will be coloured as
            zerocolor, all negative values as negcolor and all positive values as poscolor.

            In the numpy array case each connection will be set to the colour as specified in
            the same position in the array.  If the entry is None then defaultcolor will be used

        defaultlinkcolor: str, optionals
            defaults to 'black'.  Only used if linkcolors is a numpy array
            and None entries are found

        Returns
        -------
        str
            String containing links.conf content needed for the relevant
            plot configuration
        """

        s = []

        if linkcolors is None:
            # zero, neg, pos
            linkcolors = ('black', 'blue', 'red')

        num_regions = self.num_regions

        if data.shape != (num_regions, num_regions):
            raise Exception("Data must be {} x {}".format(num_regions, num_regions))

        # Generate a cache of map lookups and region IDs to names
        cache = self.circular_link_cache
        region_names = self.data_regions

        linefmt = "{} {} {} {} {} {} thickness={}p,color={}"

        for row in range(num_regions):
            for col in range(num_regions):
                val = data[row, col]

                # Skip any entries which are NaN
                if np.isnan(val):
                    continue

                if isinstance(linkcolors, (tuple, list)):
                    if val == 0:
                        color = linkcolors[0]
                    elif val < 0:
                        color = linkcolors[1]
                    else:
                        color = linkcolors[2]
                else:
                    # Assume a numpy array of strings
                    color = linkcolors[row, col]

                if color is None:
                    color = defaultlinkcolor

                reg_from = region_names[row]
                reg_to = region_names[col]

                # Note that thickness must be positive
                s.append(linefmt.format(cache[reg_from][0], cache[reg_from][1], cache[reg_from][2],
                                        cache[reg_to][0], cache[reg_to][1], cache[reg_to][2],
                                        np.abs(val), color))

        return '\n'.join(s)

    def parse_colors(self, linkcolors, defaultlinkcolor):
        """
        Parse hex colours into something usable by circos
        """

        # Map (r, g, b) tuples to colour names
        colour_dict = {}
        colidx = 0

        if isinstance(linkcolors, (tuple, list)):
            lc = []
            for colour in linkcolors:
                if colour.startswith('#'):
                    # Parse as #RGB and add to colour_dict
                    r, g, b = mc.hex2color(colour)
                    r = int(r * 255)
                    g = int(g * 255)
                    b = int(b * 255)

                    if (r, g, b) not in colour_dict.keys():
                        colour_dict[(r, g, b)] = 'customcolor{}'.format(colidx)
                        colidx += 1

                    lc.append(colour_dict[(r, g, b)])
                else:
                    # Leave as it is
                    lc.append(colour)

            lc = tuple(lc)
        elif isinstance(linkcolors, np.ndarray):
            lc = linkcolors.copy()

            for row in range(lc.shape[0]):
                for col in range(lc.shape[1]):
                    colour = lc[row, col]
                    if colour is not None and colour.startswith('#'):
                        # Parse as #RGB and add to colour_dict
                        r, g, b = mc.hex2color(colour)
                        r = int(r * 255)
                        g = int(g * 255)
                        b = int(b * 255)

                        if (r, g, b) not in colour_dict.keys():
                            colour_dict[(r, g, b)] = 'customcolor{}'.format(colidx)
                            colidx += 1

                        lc[row, col] = colour_dict[(r, g, b)]

        elif linkcolors is None:
            lc = None
        else:
            raise Exception("Cannot parse link colors {}".format(linkcolors))

        if defaultlinkcolor is None:
            dlc = None
        elif defaultlinkcolor.startswith("#"):
            # Parse as #RGB and add to colour_dict
            r, g, b = mc.hex2color(defaultlinkcolor)
            r = int(r * 255)
            g = int(g * 255)
            b = int(b * 255)

            if (r, g, b) not in colour_dict.keys():
                colour_dict[(r, g, b)] = 'customcolor{}'.format(colidx)
                colidx += 1

            dlc = colour_dict[(r, g, b)]
        else:
            dlc = defaultlinkcolor

        # Work out what to put in the colors section (if anything)
        colour_text = ""

        if len(colour_dict) > 0:
            colour_text = """<colors>\n"""

            for k, v in colour_dict.items():
                colour_text += "{} = {},{},{}\n".format(v, k[0], k[1], k[2])

            colour_text += """</colors>\n"""

        return lc, dlc, colour_text

    def gen_circos_files(self, data, karyotype_name, output_name,
                         radius=1500,
                         circos_path='/etc/circos',
                         output_dir=None, **kwargs):
        """
        Create a Circos plot.

        Note that additional kwargs parameters can be passed to this routine.
        They will be passed to the ticks() and links() routines.  See those
        routines for details of the additional parameters available.

        Parameters
        ----------
        data : ndarray
            a (N x N) matrix containing thickness (and np.nan where no connection is desired)

        karyotype_name: str
            the filename of your karyotype file on disk

        output_name: str
            output basename of configuration files (with no extension)

        radius: int, optional
            in pixels (defaults to 1500)

        circos_path: str, optional
            where to find the circos configuration files (defaults to /etc/circos)

        output_name: str, optional
            Directory in which to store output

        circle_prop: float, optional
            proportion of image to fill circle with (defaults to 0.85).  Must
            be in range 0-1.


        Note that if you are creating plots for publication using Circos, you should
        cite the relevant publication: [Krzywinski2009]_.
        """

        linksname = output_name + '.txt'

        linkspath = linksname
        if output_dir is not None:
            linkspath = join(output_dir, linksname)

        imagename = output_name + '.png'

        confpath = output_name + '.conf'
        if output_dir is not None:
            confpath = join(output_dir, confpath)

        linkcolors = kwargs.get('linkcolors', None)
        defaultlinkcolor = kwargs.get('defaultlinkcolor', 'black')

        lc, dlc, colors = self.parse_colors(linkcolors, defaultlinkcolor)

        circle_radius = kwargs.get('circle_prop', 0.85)

        # First of all, create the .txt links file
        f = open(linkspath, 'w')
        f.write(self.links(data, lc, dlc))
        f.close()

        # Now the config file - pass all keyword arguments down
        ticks = self.ticks(**kwargs)

        f = open(confpath, 'w')
        f.write(CIRCOS_CONF_TEMPLATE.format(karyotype=karyotype_name,
                                            ticks=ticks,
                                            radius=radius,
                                            colors=colors,
                                            circospath=circos_path,
                                            linksname=linksname,
                                            outputimage=imagename,
                                            circle_radius=circle_radius))
        f.close()

    def gen_circos_plot(self, data, output_name,
                        radius=1500, circos_path='/etc/circos', **kwargs):
        """
        Generate a circos plot output without leaving behind the configuration files.

        All arguments to this function are the same as gen_circos_files except
        that you do not specify a karyotype_name as this is unnecessary.

        Note that if you are creating plots for publication using Circos, you should
        cite the relevant publication: [Krzywinski2009]_.
        """
        try:
            from subprocess import run, PIPE
        except ImportError:
            warnings.warn("subprocess.run is not present; this routine is not supported on python2")
            return

        with tempfile.TemporaryDirectory() as d:
            # Generate the karyotype
            f_k = open(join(d, 'karyotype.txt'), 'w')
            f_k.write(self.karyotype())
            f_k.close()

            # Generate the config files
            self.gen_circos_files(data, 'karyotype.txt',
                                  'circos',
                                  radius=radius,
                                  circos_path=circos_path,
                                  output_dir=d,
                                  **kwargs)

            # Set up the symlinks which seem to be necessary
            makedirs(join(d, 'etc'))
            symlink(join(circos_path, 'tracks'), join(d, 'etc', 'tracks'))

            # Now run circos and check for the error code
            # Can't use check_output as we're still on python3.5 (Debian stretch)
            r = run(['circos', '-conf', 'circos.conf'], cwd=d,
                    stdout=PIPE, stderr=PIPE)

            if r.returncode != 0:
                raise CircosException(r.returncode, r.stdout, r.stderr)

            # Move files to final location
            copyfile(join(d, 'circos.png'), output_name + '.png')
            copyfile(join(d, 'circos.svg'), output_name + '.svg')

    def plot_netmat(self, data, ax=None, labels=True,
                    hregions='top', vregions='right',
                    show_xticks=True, show_yticks=True,
                    showgrid=True, **kwargs):
        """
        Plot a netmat with coloured bars to represent the grouping of the regions.

        Required arguments:
          data: a (N, N) matrix of data to be plotted

        Optional arguments:
          ax: An axes object on which to plot.  One will be created on a new figure
              if not provided.
          labels: If True, use the shortnames, if False do not plot labels,
                  if a list, use the list as the labels (must be N long)
          hregions: Location to place horizontal coloured bars indicating regions of ROIs.
                    One of 'top', 'bottom', 'off'.  Defaults to 'top'
          vregions: Location to place vertical coloured bars indicating regions of ROIs.
                    One of 'left', 'right', 'off'.  Defaults to 'right'
          show_xticks: Whether to show the xticks on the axis
          show_yticks: Whether to show the yticks on the axis
          showgrid: If True, place a grid on the image to delineate the regions.

        Keyword arguments:
          label_fontsize: will be passed to plot_netmat to set the font size

        Any additional keyword arguments will be passed down to pcolormesh to allow
        setting things such as the colormap
        """

        from sails.plotting import plot_netmat

        # We need the colours in a normal form
        colour_dict = self.colour_mapping(normalise=True)

        colour_names = []
        tick_pos = []
        count = 0

        if ax is None:
            plt.figure()
            ax = plt.subplot(1, 1, 1)

        # We need to know which colour each ROI is
        for group in self.netmat_groups:
            g = self.groups[group]
            colour_names.extend([g.circoscolour] * len(g))
            tick_pos.append(count + len(g))
            count += len(g)

        xtick_pos = None
        ytick_pos = None

        if show_xticks:
            xtick_pos = tick_pos

        if show_yticks:
            ytick_pos = tick_pos

        if labels is True:
            labels = self.netmat_shortnames
        elif labels is False:
            labels = None

        # Otherwise assume that they have provided a list of labels
        # which they want to use

        # Order data properly - can't think of a better way to do this
        idx = np.array(self.netmat_indices)
        new_data = data[idx, :][:, idx]

        return plot_netmat(new_data, ax=ax, colors=colour_dict, cnames=colour_names,
                           xtick_pos=xtick_pos, ytick_pos=ytick_pos, labels=labels,
                           hregions=hregions, vregions=vregions,
                           showgrid=showgrid, **kwargs)



#!/usr/bin/python

# vim: set expandtab ts=4 sw=4:

import unittest

import numpy as np


class MorletBasisTests(unittest.TestCase):

    def test_morlet_basis_generation(self):
        sample_rate = 128
        freqs = np.linspace(1, 50, 49)
        ncycles = 5

        from ..wavelet import get_morlet_basis

        mlt = get_morlet_basis(freqs, ncycles, sample_rate)

        # Check we have same number of morlet basis waves as input freqs
        assert(len(mlt) == freqs.shape[0])

        # Check that each morlet basis wave is of shorter or equal length to
        # previous one (for strictly ascending freqs)
        assert(np.all(np.diff([len(m) for m in mlt]) <= 0))

        # Example with faster sample_rate
        mlt2 = get_morlet_basis(freqs, ncycles, sample_rate*2)
        check = np.c_[[len(m) for m in mlt], [len(m) for m in mlt2]]

        # Morlets from double sample rate should be twice as long (or off-by-one)
        maxdiff = np.max(check[:, 0] - check[:, 1]//2)
        assert(0 <= maxdiff and maxdiff < 2)


class MorletTransformTests(unittest.TestCase):

    def setUp(self):

        self.seconds = 10
        self.sample_rate = 512
        self.time = np.linspace(0, self.seconds, self.seconds*self.sample_rate)
        self.f1 = 10
        self.f2 = 3

        self.x1 = np.sin(2 * np.pi * self.f1 * self.time[:(self.seconds // 2) * self.sample_rate])
        self.x2 = 2 * np.sin(2 * np.pi * self.f2 * self.time[(self.seconds // 2) * self.sample_rate:])

        self.x = np.r_[self.x1, self.x2]

    def test_simple_morlet_norm(self):

        from ..wavelet import morlet

        freqs = np.linspace(1, 50, 49)

        cwt = morlet(self.x, freqs, self.sample_rate,
                     ret_mode='amplitude', normalise='simple')

        # Check that amplitude is close to 'true' value whilst each oscillation
        # is 'on' and close to zerom when it is 'off'

        # Early signal, check 10Hz is close to 1 and 3Hz is close to zero
        assert(cwt[9, 2000] > .97 and cwt[9, 2000] < 1.03)
        assert(cwt[2, 2000] < 0.03)

        # Late signal, check 3Hz is close to 2 and 10Hz is close to zero
        assert(cwt[9, 4000] < 0.03)
        assert(cwt[2, 4000] > 1.97 and cwt[2, 4000] < 2.03)



#!/usr/bin/python

# vim: set expandtab ts=4 sw=4:

__all__ = []


def array_assert(a, b, decimal=None, **kwargs):
    """Function checking whether two arrays are equal in size and almost_equal
    in content (using numpy.testing.assert_almost_equal).

    Parameters
    ----------
    a : ndarray
        First array to check
    b : ndarray
        Second array to check
    decimal : float
        The precisionof the comparison (Default value = None)
    **kwargs : dict
        Extra arguments to pass to numpy.testing.assert_almost_equal

    Returns
    -------

    """
    if decimal is None:
        if a.shape != b.shape:
            raise AssertionError("Sizes of matrices don't match "
                                 "(%s vs %s)" % (str(a.shape),
                                                 str(b.shape)))

        if ((a == b).all()):
            return

        # Otherwise give some info as to why we're asserting
        raise AssertionError("Arrays are not the same:\n"
                             "Array A:%s\nArray B:%s" % (a, b))
    else:
        from numpy.testing import assert_almost_equal
        assert_almost_equal(a, b, decimal, *kwargs)


__all__.append('array_assert')


def ensure_leading_positive(A):
    """This routine ensures that any parameter matrix is of a form which has a
    leading positive identity matrix.  If this is already the case, the
    parameter matrix will be returned unchanged; if it is not, a negated
    version of the parameter matrix will be returned.

    Parameters
    ----------
    A : ndarray
        Parameter matrix (3 or 4D)

    Returns
    -------
    ndarray
        Parameter matrix where the leading identity is guaranteed to be
        positive


    """

    # We just check the first entry in the array
    if A.ndim == 3:
        entry = A[0, 0, 0]
    elif A.ndim == 4:
        entry = A[0, 0, 0, 0]
    else:
        raise ValueError("Parameter matrix must be 3- or 4D")

    # We could add a full check at this point for a leading identity and raise
    # an error if we wanted to
    if entry > 0:
        # We are already positive
        return A

    return -A


__all__.append('ensure_leading_positive')


def ensure_leading_negative(A):
    """This routine ensures that any parameter matrix is of a form which has a
    leading negative identity matrix.  If this is already the case, the
    parameter matrix will be returned unchanged; if it is not, a negated
    version of the parameter matrix will be returned.

    Parameters
    ----------
    A :
        Parameter matrix (3 or 4D)

    Returns
    -------
    ndarray
        Parameter matrix where the leading identity is guaranteed to be
        negative

    """

    # We just check the first entry in the array
    if A.ndim == 3:
        entry = A[0, 0, 0]
    elif A.ndim == 4:
        entry = A[0, 0, 0, 0]
    else:
        raise ValueError("Parameter matrix must be 3- or 4D")

    # We could add a full check at this point for a leading identity and raise
    # an error if we wanted to
    if entry < 0:
        # We are already negative
        return A

    return -A


__all__.append('ensure_leading_negative')



#!/usr/bin/python

# vim: set expandtab ts=4 sw=4:

import unittest

import numpy as np

from ..support import array_assert


class DelayDiagnosticTests(unittest.TestCase):
    def test_constructor(self):
        """Test that we can initialise a DelayDiagnostics object"""
        from ..diags import DelayDiagnostics

        d = DelayDiagnostics()

        assert(d.MI is None)
        assert(d.MI_diff is None)
        assert(d.delay_vect_samples is None)
        assert(d.delay_vect_ms is None)
        assert(d.time_vect is None)
        assert(d.autocorrelation is None)
        assert(d.maxdelay is None)
        assert(d.first_zero is None)

    def test_delay_assessment_constant(self):
        """Test that we can assess delay with a constant window (default)"""
        from ..diags import DelayDiagnostics
        # assert_raises(ValueError, mutual_information, np.zeros(1000,),
        #               np.zeros(1000, ))
        # array_assert(ret, array([1.50009220434754353413]))

        # Create a simple sine/cosine set to test with
        t = np.linspace(0, 1, 1000)
        dat = np.array([[np.sin(2 * np.pi * t * 10).T],
                        [np.cos(2 * np.pi * t * 10).T]])
        # Needs to be signals / timepts / epochs
        dat = dat.transpose(0, 2, 1)

        d = DelayDiagnostics.delay_search(dat, 100, 1, 1000.0)

        # TODO: Ideally, we'd have a better test to fully check the
        # returned values based on the analytic version.
        array_assert(d.delay_vect_samples, np.arange(1, 100))
        array_assert(d.delay_vect_ms, np.arange(1, 100) * (1/1000.0))
        # This should be at time == 25ms
        # So in samples:
        assert(d.first_zero == 24)
        assert(d.delay_vect_ms[d.first_zero] == 0.025)
        assert(d.autocorrelation.shape == (2, 1, 1000))
        assert(d.MI.shape == (2, 99))
        assert(d.MI_diff.shape == (2, 98))

    def test_delay_search_nonconstant(self):
        """Test that we can assess delay with a non-constant window"""
        from ..diags import DelayDiagnostics

        # Create a simple sine/cosine set to test with
        t = np.linspace(0, 1, 1000)
        dat = np.array([[np.sin(2*np.pi * t * 10).T],
                        [np.cos(2*np.pi * t * 10).T]])
        # Needs to be signals / timepts / epochs
        dat = dat.transpose(0, 2, 1)

        d = DelayDiagnostics.delay_search(dat, 100, 1, 1000.0, False)

        # TODO: Ideally, we'd have a better test for the non-constant window
        # code.
        array_assert(d.delay_vect_samples, np.arange(1, 100))
        array_assert(d.delay_vect_ms, np.arange(1, 100) * (1/1000.0))
        # This should be at time == 25ms
        # So in samples:
        assert(d.first_zero == 24)
        assert(d.delay_vect_ms[d.first_zero] == 0.025)
        assert(d.autocorrelation.shape == (2, 1, 1000))
        assert(d.MI.shape == (2, 99))
        assert(d.MI_diff.shape == (2, 98))



#!/usr/bin/python

# vim: set expandtab ts=4 sw=4:

import unittest

import numpy as np

__all__ = []


class test_modal_decomp(unittest.TestCase):

    def test_modal_decomp(self):
        """Test modal decomposition"""
        from ..modal import MvarModalDecomposition
        from ..modelfit import VieiraMorfLinearModel

        sample_rate = 128
        seconds = 10
        f = 20

        time_vect = np.linspace(0, seconds, seconds * sample_rate)
        X = np.sin(2 * np.pi * f * time_vect)[None, :, None]

        delay_vect = np.arange(3)
        m = VieiraMorfLinearModel.fit_model(X, delay_vect)
        mo = MvarModalDecomposition.initialise(m, sample_rate)

        # Check we have the right number of modes
        assert(mo.evals.shape[0] == 2)

        # Check peak frequency to 1 decimal place
        assert(np.round(mo.peak_frequency, 1)[0, 0] == f)



#!/usr/bin/python

# vim: set expandtab ts=4 sw=4:

from .modal import *  # noqa: F401, F403
from .modelfit import *  # noqa: F401, F403
from .modelvalidation import *  # noqa: F401, F403
from .mvar_metrics import *  # noqa: F401, F403
from .diags import *  # noqa: F401, F403
from .plotting import *  # noqa: F401, F403
from .circular_plots import *  # noqa: F401, F403
from .stats import *  # noqa: F401, F403
from .tutorial_utils import *  # noqa: F401, F403
from .simulate import *  # noqa: F401, F403
from .orthogonalise import *  # noqa: F401, F403
from .utils import *  # noqa: F401, F403
from .wavelet import *  # noqa: F401, F403



#!/usr/bin/python

# vim: set expandtab ts=4 sw=4:

import unittest

import numpy as np

from ..support import array_assert


class ResampleTests(unittest.TestCase):
    def test_fast_resample_int_factor(self):
        """Test that we properly resample data with an integer factor"""
        from ..utils import fast_resample

        # Create a 10Hz signal sampled at 500Hz for 2s
        t = np.linspace(0, 2-(1/1000), 1000)
        s = np.sin(2*np.pi*t*10)

        # Reformat to [nchannels, nsamples]
        s = s[np.newaxis, :]

        # Re-sample by a factor of 2 (to 250Hz)
        res = fast_resample(s, 2)

        # We should still have 20 peaks
        # Find them by looking for points where the derivative reverses sign
        diff_res = np.diff(res)[0, :]
        peaks = []
        for idx in range(1, diff_res.shape[0]):
            if diff_res[idx-1] > 0 and diff_res[idx] < 0:
                peaks.append(idx)

        assert(len(peaks) == 20)

        # And each peak should be 25 samples apart starting at sample 6
        assert(peaks[0] == 6)
        diff = np.array([peaks[x] - peaks[x-1] for x in range(1, len(peaks))])
        assert(np.sum(diff == 25) == 19)

    def test_fast_resample_alt_dim(self):
        """Test that we properly resample data with an alternate dimension"""
        from ..utils import fast_resample

        # Create a 10Hz signal sampled at 500Hz for 2s
        t = np.linspace(0, 2-(1/1000), 1000)
        s = np.sin(2*np.pi*t*10)

        # Re-sample by a factor of 2 (to 250Hz)
        res = fast_resample(s, 2, axis=0)

        # We should still have 20 peaks
        # Find them by looking for points where the derivative reverses sign
        diff_res = np.diff(res)
        peaks = []
        for idx in range(1, diff_res.shape[0]):
            if diff_res[idx-1] > 0 and diff_res[idx] < 0:
                peaks.append(idx)

        assert(len(peaks) == 20)

        # And each peak should be 25 samples apart starting at sample 6
        assert(peaks[0] == 6)
        diff = np.array([peaks[x] - peaks[x-1] for x in range(1, len(peaks))])
        assert(np.sum(diff == 25) == 19)

    def test_fast_resample_float_factor(self):
        """Test that we properly resample data with an float factor"""
        from ..utils import fast_resample

        # Create a 10Hz signal sampled at 500Hz for 2s
        t = np.linspace(0, 2-(1/1000), 1000)
        s = np.sin(2*np.pi*t*10)

        # Reformat to [nchannels, nsamples]
        s = s[np.newaxis, :]

        # Re-sample by a factor of 2.5 (to 200Hz)
        res = fast_resample(s, 2.5)

        # We should still have 20 peaks
        # Find them by looking for points where the derivative reverses sign
        diff_res = np.diff(res)[0, :]
        peaks = []
        for idx in range(1, diff_res.shape[0]):
            if diff_res[idx-1] > 0 and diff_res[idx] < 0:
                peaks.append(idx)

        assert(len(peaks) == 20)

        # And each peak should be 19 or 20 samples apart starting at sample 5
        # (there is some inevitable rounding error)
        assert(peaks[0] == 5)
        diff = np.array([peaks[x] - peaks[x-1] for x in range(1, len(peaks))])
        assert(np.sum((diff == 20) | (diff == 19)) == 19)

    def test_fast_resample_multi_dim(self):
        """Test that we properly resample data in multi dimension mode"""
        from ..utils import fast_resample

        # Create a 10Hz signal sampled at 500Hz for 2s
        t = np.linspace(0, 2-(1/1000), 1000)
        s = np.sin(2*np.pi*t*10)

        # Reformat to [nchannels, nsamples]
        s = s[:, np.newaxis, np.newaxis]

        s = np.tile(s, (1, 1, 2))

        # Re-sample by a factor of 2 (to 250Hz)
        res = fast_resample(s, 2, axis=0)

        # For each of our two copies, perform our tests
        for idx in range(2):
            test_data = res[:, 0, idx]
            # We should still have 20 peaks
            # Find them by looking for points where the derivative reverses sign
            diff_res = np.diff(test_data)
            peaks = []
            for idx in range(1, diff_res.shape[0]):
                if diff_res[idx-1] > 0 and diff_res[idx] < 0:
                    peaks.append(idx)

            assert(len(peaks) == 20)

            # And each peak should be 25 samples apart starting at sample 6
            assert(peaks[0] == 6)
            diff = np.array([peaks[x] - peaks[x-1] for x in range(1, len(peaks))])
            assert(np.sum(diff == 25) == 19)


class EpochingTests(unittest.TestCase):

    def test_seglen_epoching(self):
        X = np.arange(10 * 1000).reshape(10, 1000)

        from ..utils import epoch_data

        Xout = epoch_data(X, segment_len=15)
        # Check basic shape
        assert(Xout.shape[1] == 15)
        assert(Xout.shape[2] == 66)

        # Check trial lengths are as expected
        assert(np.all(np.diff(Xout[0, 0, :]) == 15))
        # Check first trial values are as expected
        assert(np.all(Xout[0, :, 0] == np.arange(15)))
        assert(np.all(Xout[0, :, 20] == np.arange(300, 315)))

    def test_trials_epoching(self):
        X = np.arange(10 * 1000).reshape(10, 1000)

        trials = np.repeat([40, 100, 150, 210, 300], 2).reshape(-1, 2)
        trials[:, 1] += 40

        from ..utils import epoch_data

        Xout = epoch_data(X, trials=trials)
        # Check basic shape
        assert(Xout.shape[1] == 40)
        assert(Xout.shape[2] == 5)

        # Check first trial values are as expected
        assert(np.all(Xout[0, :, 1] == np.arange(trials[1, 0], trials[1, 1])))


class ValidSamplesTests(unittest.TestCase):
    def test_valid_samples(self):
        """Test that we properly compute the indices for valid samples"""
        from ..utils import get_valid_samples

        # Create some test data: 2 chans, 100 samples, 10 trials
        # Set up data as (trial_num*100) + (chan_no * 2000) + (sample_num)
        num_chans = 2
        num_samples = 100
        num_trials = 10

        data = np.zeros((num_chans, num_samples, num_trials))
        for k in range(num_chans):
            for l in range(num_trials):
                data[k, :, l] = (k * 2000) + (l * 100) + np.arange(num_samples)

        # Assume that we have an order 5 model
        delay_vect = np.arange(5 + 1)

        # Check that in the 'full' case we just get back our data
        array_assert(data, get_valid_samples(data, delay_vect, mode='full', backwards=False))
        array_assert(data, get_valid_samples(data, delay_vect, mode='full', backwards=True))

        # Check that in the 'valid' forward case, we get back our data offset by 6 samples (order + 1)
        array_assert(data[:, 6:, :],
                     get_valid_samples(data, delay_vect, mode='valid', backwards=False))

        # Check that in the 'valid' backwards case, we get back our data with 6 samples taken from
        # the end
        array_assert(data[:, :-6, :],
                     get_valid_samples(data, delay_vect, mode='valid', backwards=True))

        # Check that in the 'full_nan' forward case, we get back our data where the first 6 samples
        # are NaN
        nan_data = data.copy()
        nan_data[:, 0:6, :] = np.nan

        valid = get_valid_samples(data, delay_vect, mode='full_nan', backwards=False)
        # Check NaNs
        assert(np.sum(np.isnan(valid[:, 0:6, :])) == 120)
        # Check rest of array
        array_assert(nan_data[:, 6:, :], valid[:, 6:, ])

        # Check that in the 'full_nan' backward case, we get back our data where the last 6 samples
        # are NaN
        nan_data = data.copy()
        nan_data[:, -6:, :] = np.nan

        valid = get_valid_samples(data, delay_vect, mode='full_nan', backwards=True)
        # Check NaNs
        assert(np.sum(np.isnan(valid[:, -6:, :])) == 120)
        # Check rest of array
        array_assert(nan_data[:, :-6, :], valid[:, :-6, ])

        # Check that we raise on a bad mode
        self.assertRaises(ValueError, get_valid_samples,
                          data, delay_vect, mode='notreal', backwards=True)


class ArtefactDetectionTests(unittest.TestCase):
    def test_gesd(self):
        from ..utils import gesd
        np.random.seed(43)

        # Create some random data
        X = np.random.randn(100, 1)

        # Check that we don't detect outliers in normally distributed random
        # data
        bads, cleanX = gesd(X)
        # No bad samples
        assert(np.all(bads == False))  # noqa: E712
        # Clean data is the same as data
        assert(np.all(X == cleanX))

        # and on an outlier side -1 test
        bads, cleanX = gesd(X, outlier_side=-1)
        # No bad samples
        assert(np.all(bads == False))  # noqa: E712
        # Clean data is the same as data
        assert(np.all(X == cleanX))

        # and on an outlier side 1 test
        bads, cleanX = gesd(X, outlier_side=1)
        # No bad samples
        assert(np.all(bads == False))  # noqa: E712
        # Clean data is the same as data
        assert(np.all(X == cleanX))

        # and that we deal with passing a list
        bads, cleanX = gesd(list(X))
        # No bad samples
        assert(np.all(bads == False))  # noqa: E712
        # Clean data is the same as data
        assert(np.all(X == cleanX))

        # An an obviously bad sample
        X[50] = 10
        bads, cleanX = gesd(X)
        # Check correct sample is identified
        assert(np.where(bads == True)[0][0] == 50)  # noqa: E712
        # Bad sample is identified
        assert(cleanX.shape[0] == X.shape[0]-1)

    def test_detect_artefacts(self):
        from ..utils import detect_artefacts

        # Create random data
        np.random.seed(42)
        X = np.random.randn(50, 1000, 10)

        # Check no bads detected
        assert(np.sum(detect_artefacts(X, 0)) == 0)
        assert(np.sum(detect_artefacts(X, 1)) == 0)
        assert(np.sum(detect_artefacts(X, 2)) == 0)

        X[10, :, :] *= 5

        # Check one bad detected
        assert(np.sum(detect_artefacts(X, 0)) == 1)
        # Check it it at index 10
        assert(np.where(detect_artefacts(X, 0))[0] == 10)

        # Check one bad detected
        assert(np.sum(detect_artefacts(X, 0, ret_mode='good_inds')) == 49)
        # Check it it at index 10
        assert(np.where(detect_artefacts(X, 0, ret_mode='good_inds') == False)[0][0] == 10)  # noqa: E712

        # Check it all works with dim 3
        X[:, :, 2] *= 8

        # Check one bad detected
        assert(np.sum(detect_artefacts(X, 2)) == 1)
        # Check it it at index 10
        assert(np.where(detect_artefacts(X, 2))[0] == 2)

        # Check one bad detected
        assert(np.sum(detect_artefacts(X, 2, ret_mode='good_inds')) == 9)
        # Check it it at index 10
        assert(np.where(detect_artefacts(X, 2, ret_mode='good_inds') == False)[0][0] == 2)  # noqa: E712

        # Check nan outputs are sensible
        assert(np.isnan(detect_artefacts(X, 2, ret_mode='nan_bads')).sum() == 50000)
        assert(np.isnan(detect_artefacts(X, 0, ret_mode='nan_bads')).sum() == 10000)

        # Create random data
        np.random.seed(42)
        X = np.random.randn(50, 1000, 10)
        # Add high variance epoch
        X[:, 650:750, :] *= 10

        # Check that correct number of samples are rejected
        assert(np.sum(detect_artefacts(X, 1, reject_mode='segments', segment_len=50)) == 100)
        # Check that start and end of epoch are correctly identified
        assert(detect_artefacts(X, 1, reject_mode='segments', segment_len=50)[649] == False)  # noqa: E712
        assert(detect_artefacts(X, 1, reject_mode='segments', segment_len=50)[650] == True)  # noqa: E712
        assert(detect_artefacts(X, 1, reject_mode='segments', segment_len=50)[749] == True)  # noqa: E712
        assert(detect_artefacts(X, 1, reject_mode='segments', segment_len=50)[750] == False)  # noqa: E712

    def test_bad_params(self):
        from ..utils import detect_artefacts

        np.random.seed(43)

        # Create a dataset in sails format with 32 channels, 500 samples and
        # 128 epochs.
        X = np.random.randn(32, 500, 128)

        # No axis passed
        self.assertRaises(ValueError, detect_artefacts, X)

        # Bad reject mode
        self.assertRaises(ValueError, detect_artefacts, X, reject_mode='unknown')



#!/usr/bin/python

import numpy as np
from scipy import signal

__all__ = []

# ABSTRACT CLASSES


class AbstractSigGen():
    """Abstract base class for signal generation"""

    def __init__(self, extraargs=None):
        if extraargs is not None:
            self.parse_arguments(extraargs)

    def get_num_sources(self):
        """Helper function for returning number of sources in simulation.

        This is typically overloaded when defining a signal generator class.

        Returns
        -------
        type : int
            Number of sources in simulated signal. If not override, returns 1

        """
        return 1

    def parse_arguments(self, extraargs):
        """Helper function for processing extra-arguments passed to a
        signal-generator.

        Currently only looks for snr_dB for noise levels but can be overloaded
        by a specific class to processes more arguments.

        Parameters
        ----------
        extraargs : dict
            keyword arguments passed to signal generator.

        """
        if not hasattr(self, 'snr_dB'):
            try:
                self.snr_dB = float(extraargs.get('snr_dB', 0.))
            except (TypeError, KeyError):
                raise TypeError(" SNR is not defined (use snr_dB argument)")

    def generate_signal(self, meg_reader):  # pragma: no cover
        """Generate a custom signal for embedding in a dataset

        Parameters
        ----------
        meg_reader :
            MEG data for which to generate signal

        Returns
        -------
        type
            numpy array of shape (samples).  Note that samples must
            be equal to the value of meg_reader.num_samples

        """

        raise NotImplementedError("This is an abstract base class")


__all__.append('AbstractSigGen')


class AbstractLagSigGen(AbstractSigGen):
    """Generic class for making lagged signals"""

    def get_num_sources(self):
        """ """
        print("Warning! get_num_sources is not implemented")
        return None

    def parse_arguments(self, extraargs):
        """Helper function for processing extra-arguments passed to a
        signal-generator.

        Currently only looks for snr_dB for noise levels but can be overloaded
        by a specific class to processes more arguments.

        Parameters
        ----------
        extraargs : dict
            keyword arguments passed to signal generator.

        """

        # Make sure we call the parents parse args as well to get generic
        # parameters such as snr_dB
        super(AbstractMVARSigGen, self).parse_arguments(extraargs)

    def generate_basesignal(self, nsamples):
        """Function defining the starting signal which will be lagged throughout
        the network

        This is an abstracted function and should be overloaded when creating a
        new class.

        Parameters
        ----------
        nsamples : int
            The number of samples to generate

        """
        raise NotImplementedError("This is an abstract base class, "
                                  "generate_basesignal is not defined")

    def generate_noise(self, nsamples):
        """Function defining the additive noise added at each lag

        Parameters
        ----------
        nsamples : int
            The number of samples to generate

        Returns
        -------
        ndarray
            A vector of white noise

        """
        return np.random.randn(nsamples)

    def get_lags(self):
        """Method returning the lagging of the signal through the system. Should

        Returns
        -------
        type
            dimension refers to the source node and the second dimension the target
            node. Each non-zero value indicates that the signal in the first
            dimension will be added into the signal in the second dimension at the
            specified weight.

        """
        raise NotImplementedError("This is an abstract base class, "
                                  "generate_parameters is not defined")

    def generate_signal(self, sample_rate=None, num_samples=None,
                        noise_dB=None):
        """Method for generating a set of time courses based on either the
        structure of a meg_reader object or passed parameters and the
        self.generate_parameters method defined in this class

        Returns a timeseries of dimension [samples x signals]

        Parameters
        ----------
        sample_rate : float
             The sample rate to generate the signal at (Default value = None)
        num_samples :
             The number of samples to generate (Default value = None)
        noise_dB : float
             Noise level of the additive noise (Default value = None)

        Returns
        -------
        ndarray
            A simulated time-series [nchannels x nsamples x nrealisations]

        """

        # Create signal to be lagged
        basesignal = self.generate_basesignal(num_samples)

        # Get the lag matrix
        lags = self.get_lags()

        order = lags.shape[2]

        X = np.zeros((self.get_num_sources(), num_samples))
        X[0, :] = basesignal

        # Main Loop
        for p in range(order):
            for ii in range(self.get_num_sources()):
                for jj in range(self.get_num_sources()):
                    if lags[ii, jj, p] > 0.:
                        X[jj, p+1:] += lags[ii, jj, p] * X[ii, :-p-1] + \
                                       self.generate_noise(num_samples-1-p)

        for ii in range(self.get_num_sources()):
            X[ii, :] = (X[ii, :] - X[ii, :].mean()) / X[ii, :].std()

        return X[..., None]


__all__.append('AbstractLagSigGen')


class AbstractMVARSigGen(AbstractSigGen):
    """Generic class for making MVAR signals"""

    def get_num_sources(self):
        """ """
        print("Warning! get_num_sources is not implemented")
        return None

    def parse_arguments(self, extraargs):
        """Helper function for processing extra-arguments passed to a
        signal-generator.

        Currently only looks for snr_dB for noise levels but can be overloaded
        by a specific class to processes more arguments.

        Parameters
        ----------
        extraargs : dict
            keyword arguments passed to signal generator.

        """

        # Make sure we call the parents parse args as well to get generic
        # parameters such as snr_dB
        super(AbstractMVARSigGen, self).parse_arguments(extraargs)

    def generate_parameters(self):
        """Method to generate the parameters of the MVAR system. Should return a
        [nsignals x nsignals x nlags] parameter matrix in which the first
        dimension refers to the source node and the second dimension the target
        node

        This is an abstract method and should be overloaded when creating a
        specific class.

        """
        raise NotImplementedError("This is an abstract base class, "
                                  "generate_parameters is not defined")

    def check_stability(self):
        """Checks the stability of the parameter matrix by the magnitude of the
        largest eigenvalue of the first lag in the parameter matrix (see
        [Lutkephol2006]_ chapter 1)

        Returns
        -------
        float
            The stability index of the simulated signal

        Notes
        -----
        The stability index of a simulation should be strictly less than 1. If
        this exceeds 1 then the simulated signal is likely to tend to infinity
        or zero.

        """

        from sails.stats import stability_index

        return stability_index(self.generate_parameters())

    def generate_signal(self, sample_rate=None, num_samples=None,
                        noise_dB=None, num_realisations=1):
        """Method for generating a set of time courses based on either the
        structure of a meg_reader object or passed parameters and the
        self.generate_parameters method defined in this class

        Returns a timeseries of dimension [nchannels x nsamples x nrealisations]

        Parameters
        ----------
        sample_rate : float
             The sample rate of the simulated data (Default value = None)
        num_samples : int
             The number of samples to generate (Default value = None)
        noise_dB : float
             The level of additive noise to add to the simulationo (Default value = None)
        num_realisations : int
             Number of realisations to generate (Default value = 1)

        Returns
        -------
        ndarray
            The simulated dataset of size [nchannels x nsamples x nrealisations]

        """
        num_sources = self.get_num_sources()

        # Generate the parameters
        params = self.generate_parameters()

        order = params.shape[2]

        # Preallocate output
        X = np.zeros((num_sources, num_samples, num_realisations))

        for ep in range(num_realisations):

            # Create driving noise signal
            X[:, :, ep] = np.random.randn(num_sources, num_samples)

            # Main Loop
            for t in range(order, num_samples):
                for p in range(1, order):
                    X[:, t, ep] -= params[:, :, p].dot(X[:, t-p, ep])

        return X

    def generate_model(self):
        """Returns a LinearModel containing containing the true model parameters."""

        # We can use the AbstractLinearModel for this as we
        # don't need the fit_model routine
        from sails import AbstractLinearModel

        ret = AbstractLinearModel()
        ret.parameters = self.generate_parameters()
        ret.resid_cov = np.ones((self.get_num_sources(),
                                 self.get_num_sources(), 1))
        ret.delay_vect = np.arange(ret.order + 1)

        return ret


__all__.append('AbstractMVARSigGen')


#################################
# Literature MVAR Data Generators

class Baccala2001_fig1(AbstractMVARSigGen):
    """This class implements [Baccala2001]_ figure 1."""

    def get_num_sources(self):
        """Return the number of sources in the simulation"""

        return 3

    def generate_parameters(self):
        """Create the autoregressive parameters for [Baccala2001]_ figure 1."""

        order = 1

        X = np.zeros((self.get_num_sources(), self.get_num_sources(), order))

        X[0, :, 0] = [.5, .3, .4]
        X[1, :, 0] = [-.5, .3, .1]
        X[2, :, 0] = [0, -.3, -.2]

        X = np.concatenate((np.eye(self.get_num_sources())[..., None], -X),
                           axis=-1)

        return X


__all__.append('Baccala2001_fig1')


class Baccala2001_fig2(AbstractMVARSigGen):
    """This class implements [Baccala2001]_ figure 2."""

    def get_num_sources(self):
        """Return the number of sources in the simulation"""

        return 5

    def generate_parameters(self):
        """Create the autoregressive parameters for [Baccala2001]_ figure 2."""

        order = 3

        X = np.zeros((self.get_num_sources(), self.get_num_sources(), order))

        X[0, :, 0] = [0.95*np.sqrt(2), 0, 0, 0, 0]
        X[1, :, 0] = [0, 0, 0, 0, 0]
        X[2, :, 0] = [0, 0, 0, 0, 0]
        X[3, :, 0] = [0, 0, 0, .25*np.sqrt(2), .25*np.sqrt(2)]
        X[4, :, 0] = [0, 0, 0, .25*np.sqrt(2), .25*np.sqrt(2)]

        X[0, :, 1] = [-0.9025, 0, 0, 0, 0]
        X[1, :, 1] = [.5, 0, 0, 0, 0]
        X[2, :, 1] = [0, 0, 0, 0, 0]
        X[3, :, 1] = [-.5, 0, 0, 0, 0]
        X[4, :, 1] = [0, 0, 0, 0, 0]

        X[0, :, 2] = [0, 0, 0, 0, 0]
        X[1, :, 2] = [0, 0, 0, 0, 0]
        X[2, :, 2] = [-.4, 0, 0, 0, 0]
        X[3, :, 2] = [0, 0, 0, 0, 0]
        X[4, :, 2] = [0, 0, 0, 0, 0]

        X = np.concatenate((np.eye(self.get_num_sources())[..., None], -X),
                           axis=-1)

        return X


__all__.append('Baccala2001_fig2')


class Baccala2001_fig3(AbstractMVARSigGen):
    """This class implements [Baccala2001]_ figure 3."""

    def get_num_sources(self):
        """Return the number of sources in the simulation"""
        return 5

    def generate_parameters(self):
        """Create the autoregressive parameters for [Baccala2001]_ figure 3."""

        order = 2

        X = np.zeros((self.get_num_sources(), self.get_num_sources(), order))

        X[0, :, 0] = [0.95*np.sqrt(2), 0, 0, 0, 0]
        X[1, :, 0] = [-.5, 0, 0, 0, 0]
        X[2, :, 0] = [0, 0, 0, 0, 0]
        X[3, :, 0] = [0, 0, -.5, .25*np.sqrt(2), .25*np.sqrt(2)]
        X[4, :, 0] = [0, 0, 0, -.25*np.sqrt(2), .25*np.sqrt(2)]

        X[0, :, 1] = [-0.9025, 0, 0, 0, 0]
        X[1, :, 1] = [0, 0, 0, 0, 0]
        X[2, :, 1] = [0, .4, 0, 0, 0]
        X[3, :, 1] = [0, 0, 0, 0, 0]
        X[4, :, 1] = [0, 0, 0, 0, 0]

        X = np.concatenate((np.eye(self.get_num_sources())[..., None], -X),
                           axis=-1)

        return X


__all__.append('Baccala2001_fig3')


class Baccala2001_fig4(AbstractMVARSigGen):
    """This class implements [Baccala2001]_ figure 4."""

    def get_num_sources(self):
        """Return the number of sources in the simulation"""
        return 5

    def generate_parameters(self):
        """Create the autoregressive parameters for [Baccala2001]_ figure 4."""

        order = 2

        X = np.zeros((self.get_num_sources(), self.get_num_sources(), order))

        X[0, :, 0] = [0.95*np.sqrt(2), 0, 0, 0, 0]
        X[1, :, 0] = [-.5, 0, 0, 0, 0]
        X[2, :, 0] = [0, 0, 0, 0, 0]
        X[3, :, 0] = [0, 0, -.5, .25*np.sqrt(2), .25*np.sqrt(2)]
        X[4, :, 0] = [0, 0, 0, -.25*np.sqrt(2), .25*np.sqrt(2)]

        X[0, :, 1] = [-0.9025, 0, 0, 0, .5]
        X[1, :, 1] = [0, 0, 0, 0, 0]
        X[2, :, 1] = [0, .4, 0, 0, 0]
        X[3, :, 1] = [0, 0, 0, 0, 0]
        X[4, :, 1] = [0, 0, 0, 0, 0]

        X = np.concatenate((np.eye(self.get_num_sources())[..., None], -X),
                           axis=-1)

        return X


__all__.append('Baccala2001_fig4')


class Baccala2001_fig5(AbstractMVARSigGen):
    """This class implements [Baccala2001]_ figure 5."""

    def get_num_sources(self):
        """Return the number of sources in the simulation"""
        return 5

    def generate_parameters(self):
        """Create the autoregressive parameters for [Baccala2001]_ figure 5."""

        order = 4

        X = np.zeros((self.get_num_sources(), self.get_num_sources(), order))

        # (n - 1)
        X[0, :, 0] = [0.95*np.sqrt(2), 0, 0, 0, 0]
        X[1, :, 0] = [-.5, 0, 0, 0, 0]
        X[2, :, 0] = [0, 0, 0, 0, 0]
        X[3, :, 0] = [0, 0, -.5, .25*np.sqrt(2), .25*np.sqrt(2)]
        X[4, :, 0] = [0, 0, 0, -.25*np.sqrt(2), .25*np.sqrt(2)]

        # (n - 2)
        X[0, :, 1] = [-0.9025, 0, 0, 0, 0]
        X[1, :, 1] = [0, 0, 0, 0, 0]
        X[2, :, 1] = [0, -.4, 0, 0, 0]
        X[3, :, 1] = [0, 0, 0, 0, 0]
        X[4, :, 1] = [0, 0, 0, 0, 0]

        # Nothing at lag (n - 3) [:, :, 2]

        # (n - 4)
        X[0, :, 3] = [0, 0, 0, 0, 0]
        X[1, :, 3] = [0, 0, 0, 0, 0]
        X[2, :, 3] = [.1, 0, 0, 0, 0]
        X[3, :, 3] = [0, 0, 0, 0, 0]
        X[4, :, 3] = [0, 0, 0, 0, 0]

        X = np.concatenate((np.eye(self.get_num_sources())[..., None], -X),
                           axis=-1)

        return X


__all__.append('Baccala2001_fig5')


class Fasoula2013_eqn26(AbstractMVARSigGen):
    """This class implements [Fasoula2013]_ equation 26."""

    def get_num_sources(self):
        """Return the number of sources in the simulation"""
        return 10

    def generate_parameters(self):
        """Create the autoregressive parameters for [Fasoula2013]_ eqn 26."""

        order = 4

        X = np.zeros((self.get_num_sources(), self.get_num_sources(), order))

        X[0, :, 0] = [0.95*np.sqrt(2), 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[1, :, 0] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[2, :, 0] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[3, :, 0] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[4, :, 0] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[5, :, 0] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[6, :, 0] = [0.4, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[7, :, 0] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[8, :, 0] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[9, :, 0] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

        X[0, :, 1] = [-0.9025, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[1, :, 1] = [.5, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[2, :, 1] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[3, :, 1] = [-.5, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[4, :, 1] = [0, 0, 0, 0, 0, 0, 0, -.4, 0, 0]
        X[5, :, 1] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[6, :, 1] = [0, 0, 0, 0, 0, 0, -.4, 0, 0, 0]
        X[7, :, 1] = [0, 0, 0, 0, 0, 0, -.9, 0, 0, 0]
        X[8, :, 1] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[9, :, 1] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

        X[0, :, 2] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[1, :, 2] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[2, :, 2] = [0, .9, 0, 0, 0, 0, 0, 0, 0, 0]
        X[3, :, 2] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[4, :, 2] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[5, :, 2] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[6, :, 2] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[7, :, 2] = [0, 0, 0, 0, 0, 0, 0, .4, .3, 0]
        X[8, :, 2] = [0, 0, 0, 0, 0, 0, 0, -.3, .4, 0]
        X[9, :, 2] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

        X[0, :, 3] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[1, :, 3] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[2, :, 3] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[3, :, 3] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[4, :, 3] = [0, 0, 0, .8, 0, 0, 0, 0, 0, 0]
        X[5, :, 3] = [0, 0, 0, -.8, 0, 0, 0, 0, 0, 0]
        X[6, :, 3] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[7, :, 3] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[8, :, 3] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        X[9, :, 3] = [0, 0, 0, 0, 0, 0, 0.75, 0, 0, 0]

        return X


__all__.append('Fasoula2013_eqn26')


class Schelter2006_fig1(AbstractMVARSigGen):
    """This class implements [Schelter2006]_ figure 1."""

    def get_num_sources(self):
        """Return the number of sources in the simulation"""
        return 5

    def generate_parameters(self):
        """Create the autoregressive parameters for [Schelter2006]_ fig 1."""

        order = 4

        X = np.zeros((self.get_num_sources(), self.get_num_sources(), order))

        X[0, :, 0] = [0.6, 0, 0, 0, 0]
        X[1, :, 0] = [0, .5, 0, 0.6, 0]
        X[2, :, 0] = [0, 0, 0.8, 0, 0]
        X[3, :, 0] = [0, 0, 0, .5, 0]
        X[4, :, 0] = [0, 0, -.2, 0, .7]

        X[0, :, 1] = [0, 0.65, 0, 0, 0]
        X[1, :, 1] = [0, -.3, 0, 0, 0]
        X[2, :, 1] = [0, 0, -.7, 0, 0]
        X[3, :, 1] = [0, 0, .9, 0, .4]
        X[4, :, 1] = [0, 0, 0, 0, -.5]

        X[0, :, 2] = [0, 0, 0, 0, 0]
        X[1, :, 2] = [0, 0, 0, 0, 0]
        X[2, :, 2] = [0, 0, 0, 0, -.1]
        X[3, :, 2] = [0, 0, 0, 0, 0]
        X[4, :, 2] = [0, 0, 0, 0, 0]

        X[0, :, 3] = [0, 0, 0, 0, 0]
        X[1, :, 3] = [0, 0, -.3, 0, 0]
        X[2, :, 3] = [0, 0, 0, 0, 0]
        X[3, :, 3] = [0, 0, 0, 0, 0]
        X[4, :, 3] = [0, 0, 0, 0, 0]

        X = np.concatenate((np.eye(self.get_num_sources())[..., None],  -X),
                           axis=-1)

        return X


__all__.append('Schelter2006_fig1')


class PascualMarqui2014_fig3(AbstractMVARSigGen):
    """This class implements [Pascual-Marqui2014]_ figure 3."""

    def get_num_sources(self):
        """Return the number of sources in the simulation"""
        return 5

    def generate_parameters(self):
        """Create the autoregressive parameters for [Pascual-Marqui2014]_ fig 3."""

        order = 2

        X = np.zeros((self.get_num_sources(), self.get_num_sources(), order))

        X[0, 0, 0] = 1.5
        X[1, 0, 0] = -.2
        X[0, 1, 0] = -.25

        X[1, 1, 0] = 1.8
        X[2, 1, 0] = .9
        X[3, 1, 0] = .9
        X[4, 1, 0] = .9

        X[2, 2, 0] = 1.65
        X[3, 3, 0] = 1.65
        X[4, 4, 0] = 1.65

        X[0, 0, 1] = -.95
        X[1, 1, 1] = -.96
        X[2, 2, 1] = -.95
        X[3, 3, 1] = -.95
        X[4, 4, 1] = -.95

        X[2, 1, 1] = -.8
        X[3, 1, 1] = -.8
        X[4, 1, 1] = -.8

        X = np.concatenate((np.eye(self.get_num_sources())[..., None], -X),
                           axis=-1)

        return X


__all__.append('PascualMarqui2014_fig3')


################################
# Literature Lag Data Generators

class Korzeniewska2003Lag(AbstractLagSigGen):
    """Generate realisations of the time-lagged network defined in
        [Korzeniewska2003]_ figure 1.

    """

    def get_num_sources(self):
        """Return the number of sources in the simulation"""

        return 8

    def generate_basesignal(self, nsamples):
        """Generate oscillatory signal to be propagated through the network.
        This creates a single resonance at one-half of the Nyquist
        frequency via direct pole placement.

        Note: the original base-signal in the paper is a realisation from
        an AR model. I couldn't find the parameters for this mode, so use
        pole placement here. The spectral profile will be different to the
        paper, but the connectivity patterns are the same.

        Parameters
        ----------
        nsamples : int
            The number of samples to generate

        Returns
        -------
        ndarray
            Vector containing generated base signal.

        """

        roots = np.array((1j*.8, -1j*.8))
        b = np.poly(roots)

        return signal.filtfilt(1, b, self.generate_noise(nsamples))

    def get_lags(self):
        """Return the lagged influences between time-series in the simulated
        defined in [Korzeniewska2003]_ figure 1."""

        L = np.zeros((8, 8, 3))

        # Lag 1
        L[0, 3, 0] = 1
        L[0, 7, 0] = 1

        # Lag 2
        L[3, 1, 1] = 1
        L[7, 4, 1] = 1
        L[7, 2, 1] = 1

        # Lag 3
        L[2, 5, 2] = 1
        L[2, 6, 2] = 1

        return L


__all__.append('Korzeniewska2003Lag')


#######################
# Paper Data Generators

class SailsTutorialExample(AbstractSigGen):
    """ """

    def get_num_sources(self):
        """Return the number of sources in the simulation"""

        return 10

    def generate_basesignal(self, f, r, sample_rate, num_samples):
        """Generate a simple signal by pole placement

        Parameters
        ----------
        f : float
            The peak frequency of the resonance to simulate
        r : float ( 0 < r < 1 )
            The pole magnitude of the resonance to simulate
        sample_rate : float
            The sample rate to generate data at
        num_samples : int
            The number of data samples to generate

        Returns
        -------
        ndarray
            A vector containing a simulated resonance

        Notes
        -----
        The parameter r controls both the sharpness and magnitude of the
        generated resonance. Values close to 1 will generate exponentially
        sharper and larger peaks.

        """

        if f > 0:
            wr = (2 * np.pi * f) / sample_rate
            a1 = np.array([1, -2*r*np.cos(wr), (r**2)])
        else:
            a1 = np.poly(.75)

        return signal.filtfilt(1, a1, np.random.randn(1, num_samples)).T

    def generate_signal(self, f1, f2, sample_rate, num_samples,
                        noise_dB=None, num_realisations=1,
                        magnitude_noise=.02):
        """Method for generating a set of time courses based on either the
        structure of a meg_reader object or passed parameters and the
        self.generate_parameters method defined in this class

        Returns a timeseries of dimension [samples x signals x realisations]

        Parameters
        ----------
        f1 : float
            The frequency of the first signal
        f2 : float
            The frequency of the second  signal
        sample_rate : float
            The sample rate of the simulated signal
        num_samples : int
            The number of samples to simulate
        num_realisations : int
            The number of realisations to simulate (Default value = 1)
        magnitude_noise : float
            The variance of additive white noise (Default value = .02)

        Returns
        -------
        ndarray
            The simulated signal of size [nchannels x num_samples x num_realisations]

        """

        X = np.random.randn(self.get_num_sources(), num_samples, num_realisations)

        weight1, weight2 = self.get_connection_weights()
        v = np.random.randn()*magnitude_noise
        sig1 = self.generate_basesignal(f1, .8+v, sample_rate, num_samples)
        sig2 = self.generate_basesignal(f2, .61, sample_rate, num_samples)

        sig1 = (sig1 - sig1.mean(axis=0)) / sig1.std(axis=0)
        sig2 = (sig2 - sig2.mean(axis=0)) / sig2.std(axis=0)

        for ii in range(len(weight1)):
            X[ii, :, :] = X[ii, :, :] + weight1[ii] * np.tile(sig1, num_realisations)

        for ii in range(len(weight2)):
            X[ii, :, :] = X[ii, :, :] + weight2[ii] * np.tile(sig2, num_realisations)

        return X + np.random.randn(*X.shape)

    def get_connection_weights(self, form='vector'):
        """Define the connectivity within the simulated system.

        Parameters
        ----------
        form : {'vector' || 'matrix'}
             Which form of to return the connections  is (Default value = 'vector')

        Returns
        -------
        weight1
            vector or matrix defining connectivity for resonance 1

        weight2
            vector or matrix defining connectivity for resonance 2

        """

        weight1 = np.array([1, 1, 1, 0, 0, 0, 1, 1, 1, 0])
        weight2 = np.array([0, 0, 0, .5, .7, .9, 1, .9, .7, .5])

        if form == 'vector':
            return weight1, weight2
        elif form == 'matrix':
            return weight1[:, None].dot(weight1[None, :]), \
                   weight2[:, None].dot(weight2[None, :])


__all__.append('SailsTutorialExample')


def single_pole_oscillator(freq, sample_rate, seconds, mixing=None, r=.95, noise_std=None, random_seed=None):
    """
    Create a simulated oscillation using an autoregressive filter. A simple
    filter is defined by direct pole placement and applied to white noise to
    generate a random signal with a defined oscillatory peak frequency that
    exhibits random variability frequency, amplitude and waveform.

    Parameters
    ----------
    freq : float
        Peak frequency of signal in Hz
    sample_rate : float
        Sampling rate of the data to be simulated
    seconds : float
        Seconds of data to generate
    mixing :  1d array_like
        Weighting of signal across channels (optional - default is single channel with weight of 1)
    r : scalar
        Magnitude of pole to be created  0 >= r > 1, larger values give bigger
        oscillations with sharper spectral peaks
    noise_std : scalar
        optional scaling of white noise to be added to  simulated data
    random_seed : int
        optional setting to fix random_seed for reproducibility

    Returns
    -------
    ndarray
        data array [nchannels x nsamples]

    """

    if random_seed is not None:
        np.random.seed(random_seed)

    if mixing is None:
        mixing = np.ones((1, ))
    mixing = np.array(mixing)

    if mixing.ndim != 1:
        raise ValueError('Please input a 1d (vector) set of mixing weights')

    if freq > 0:
        freq_rads = (2 * np.pi * freq) / sample_rate
        a1 = np.array([1, -2*r*np.cos(freq_rads), (r**2)])
    else:
        a1 = np.poly(r)

    num_samples = int(sample_rate * seconds)

    x = signal.filtfilt(1, a1, np.random.randn(num_samples, ), axis=0)

    y = np.zeros((len(mixing), x.shape[0]))
    for ii in range(len(mixing)):
        y[ii, :] = mixing[ii] * x

    if noise_std is not None:
        noise = np.std(y) * noise_std * np.random.randn(num_samples, )
        y = y + noise

    if random_seed is not None:
        np.random.seed()  # restore defaults

    return y



#!/usr/bin/python

# vim: set expandtab ts=4 sw=4:

import numpy as np

from .anam import AbstractAnam, register_class
from .utils import get_valid_samples
from .stats import find_cov_multitrial, durbin_watson, \
    stability_index, stability_ratio, \
    rsquared_from_residuals, percent_consistency

from .modelvalidation import approx_model_selection_criteria

__all__ = []


class DelayDiagnostics(AbstractAnam):
    """Class which computes the mutual information as a function of lag from zero
    lag to the first zero crossing of the autocorrelation function.
    """

    # This is only used within Anamnesis
    hdf5_outputs = ['MI', 'MI_diff', 'delay_vect_samples', 'delay_vect_ms',
                    'time_vect', 'autocorrelation', 'maxdelay', 'first_zero']

    def __init__(self):
        AbstractAnam.__init__(self)

        # TODO: Add docstrings for these
        self.MI = None

        self.MI_diff = None

        self.delay_vect_samples = None

        self.delay_vect_ms = None

        self.time_vect = None

        self.autocorrelation = None

        self.maxdelay = None

        self.first_zero = None

    @classmethod
    def delay_search(cls, data, maxdelay, step, sample_rate,
                     constant_window=True):
        """Compute MI as a function of lag from zero lag to the first zero
        crossing of the autocorrelation function

        Parameters
        ----------
        data : numpy.ndarray
            array of [signals, samples, trials]
        maxdelay : int
            maximum delay to consider
        step : int
            step to increment the delay by
        sample_rate : float
            sample rate of data
        constant_window : bool
            Flag indicating that the same number of
            datapoints should be included at each delay.
            Default is True

        Returns
        -------
        DelayDiagnostics
            Populated object containing diagnostics for each value in delay

        """

        ret = cls()

        # Set up input parameters
        sample_rate = float(sample_rate)
        ret.maxdelay = maxdelay
        ret.time_vect = np.arange(data.shape[1]) * (1 / sample_rate)
        ret.delay_vect_samples = np.arange(1, ret.maxdelay)
        ret.delay_vect_ms = ret.delay_vect_samples * (1/sample_rate)

        # TODO: Check that we're not out of range of number of samples
        nsignals = data.shape[0]
        nsamples = data.shape[1]
        ntrials = data.shape[2]

        # Compute Autocorrelation
        # printv("Computing Autocorrelations")
        ac = np.zeros((nsignals, ntrials, nsamples), dtype=complex)

        for i in range(nsignals):
            for j in range(ntrials):
                ac[i, j, :] = np.correlate(data[i, :, j], data[i, :, j],
                                           'full')[nsamples-1:]

        ret.autocorrelation = ac

        # Find first zero crossing
        avg = ret.autocorrelation.mean(axis=0).mean(axis=0)
        zero_crossings = np.where(np.diff(np.sign(avg)))[0]

        ret.first_zero = zero_crossings[0]

        # printv("First zero crossing at %s samples %s ms" % (ret.first_zero,
        #                                     ret.time_vect[ret.first_zero]*1000))

        # Compute Mutual Information

        # TODO: Convert this to use stats.mutual_information
        nbins = 21
        bins = np.linspace(-8, 8, nbins)

        ret.MI = np.zeros((nsignals,
                           ret.delay_vect_samples.shape[0]),
                          dtype='float64')

        for d in range(ret.delay_vect_samples.shape[0]):

            delay = ret.delay_vect_samples[d]

            idx = nsamples - delay

            # Calculate MI for each signal, for each trial
            for s in range(nsignals):
                for t in range(ntrials):
                    if constant_window:
                        A = data[s, :-ret.maxdelay, t]
                        B = data[s, delay:-ret.maxdelay+delay, t]
                    else:
                        A = data[s, :-(delay), t]
                        B = data[s, delay:, t]

                    Ps = np.histogram(A, bins=bins)[0]
                    Pst = np.histogram(B, bins=bins)[0]
                    Ps_st = np.histogram2d(A, B, bins=bins)[0]

                    Ps = Ps / float(idx)
                    Pst = Pst / float(idx)
                    Ps_st = Ps_st / float(idx)

                    for i in range(nbins-1):
                        for j in range(nbins-1):
                            if Ps_st[i, j] != 0.0:
                                ret.MI[s, d] += Ps_st[i, j] * \
                                    np.log2(Ps_st[i, j] / (Ps[i] * Pst[j]))

        ret.MI = ret.MI / float(ntrials)

        ret.MI_diff = np.diff(ret.MI)

        # self.MI_dim_selection = model.find_elbow(self.MI)

        return ret


__all__.append('DelayDiagnostics')
register_class(DelayDiagnostics)


class ModelDiagnostics(AbstractAnam):
    """
    Class to store and display LinearModel diagnostic information. Several
    diagnostic criteria are computed including:

    R_square (``R_square``)
        the percentage of variance explained in each channel

    Stability Index (``SI``)
        Indicator of the stability of the MVAR parameters (SI<1 indicates a
        stable model)

    Stability Ratio (``SR``)
        A stronger test of stability computed from the ratio of the largest
        eigenvalue of A to all others.

    Durbin-Watson (``DW``)
        A test of autocorrelation in residuals of the model fit.  Values should
        be close to 2 indicating no autocorrelation, values close to 0 indicate
        a positive autocorrelation and 4 and negative autocorrelation.

    Log Likelihood (``LL``)
        The log-likelihood of the model.

    Akaike's Information Criterion (``AIC``)
        An indication of the model 'quality', lower values indicate a more
        accurate, less complex model.

    Bayesian Information Criterion (``BIC``)
        An indication of the model 'quality', lower values indicate a more
        accurate, less complex model.

    Percent Consistency (``PC``)
        Indicates how well a model captures the auto and cross correlation in a
        time-series.  Only computed if ``compute_pc`` is passed to the relevant
        function.
    """

    # This is only used within Anamnesis
    hdf5_outputs = ['R_square', 'SR', 'DW', 'PC', 'AIC', 'BIC', 'LL', 'SI']

    def __init__(self):
        """Constructor for ModelDiagnostics.

        This fills the diagnostic variables with placeholders. The diagnostics
        should be computed using the ModelDiagnostics.compute classmethod.

        """

        AbstractAnam.__init__(self)

        self.R_square = None
        self.SR = None
        self.DW = None
        self.PC = None
        self.AIC = None
        self.BIC = None
        self.LL = None
        self.SI = None

        self.is_list = False

    @classmethod
    def compute(cls, model, data, compute_pc=False):
        """
        Classmethod for computing a set of model diagnostics from a fitted
        model applied to a time-series dataset.

        Parameters
        ----------
        model: sails LinearModel class
            A fitted linear model
        data: ndarray
            A 3d time-series of size [nchannels x nsamples x ntrials]
        compute_pc: bool
            Flag indicating whether to compute the percent consistency, this
            can be time-consuming for large datasets (Default=False).

        Returns
        -------
        sails ModelDiagnostics instance

        """

        ret = cls()

        # Get residuals - adjust valid samples next
        resid = model.get_residuals(data, mode='full')

        # Adjust data and resids for valid samples
        data = get_valid_samples(data, model.delay_vect)
        resid = get_valid_samples(resid, model.delay_vect)

        # Get covariance of the residuals
        ret.resid_cov = find_cov_multitrial(resid, resid)

        # Compute Durbin-Watson test for residual autocorrelation
        ret.DW = durbin_watson(resid, step=np.diff(model.delay_vect)[0])[0]

        # Compute the stability index and ratio
        ret.SI = stability_index(model.parameters)
        ret.SR = stability_ratio(model.parameters)

        # Estimate the R^2 - per signal
        ret.R_square = rsquared_from_residuals(data, resid, per_signal=True)

        # Estimate Percent Consistency - Ding et al 2000.
        if compute_pc:
            # This can cause memory issues with large datasets
            ret.PC = percent_consistency(data, resid)
        else:
            ret.PC = None

        # Compute LL, AIC and BIC
        # LL, AIC, BIC = est_model_selection_criteria(model.parameters, resid)

        # Approximate AIC and BIC
        LL, AIC, BIC = approx_model_selection_criteria(model.parameters, resid)

        ret.LL = LL
        ret.AIC = AIC
        ret.BIC = BIC

        return ret

    @classmethod
    def combine_diag_list(cls, diags):
        """
        Helper function for combining diagnostics from a list of
        ModelDiagnostics instances for easy comparison and visualisation.

        Parameters
        ----------
        diags: list of ModelDiagnostics instances
            The ModelDiagnostics to concatenate

        Returns
        -------
        sails ModelDiagnostics instance

        """

        ret = cls()

        ret.resid_cov = np.concatenate([d.resid_cov[..., None] for d in diags], axis=2)

        ret.DW = np.r_[[d.DW for d in diags]]
        ret.SI = np.r_[[d.SI for d in diags]]
        ret.SR = np.r_[[d.SR for d in diags]]

        if diags[0].PC is not None:
            ret.PC = np.r_[[d.PC for d in diags]]

        ret.R_square = np.c_[[d.R_square for d in diags]]

        ret.LL = np.r_[[d.LL for d in diags]]
        ret.AIC = np.r_[[d.AIC for d in diags]]
        ret.BIC = np.r_[[d.BIC for d in diags]]

        ret.is_list = True

        return ret

    def summary(self, all_models=True):
        """Print the ModelDiagnostics in a pre-formatted table"""

        if self.is_list:
            template = "{0:<10}{1:<10}{2:<10}{3:<10}{4:<10}{5:<10}{6:<10}"

            print(template.format('Model', 'SI', 'SR', 'DW', 'AIC', 'BIC', 'R Square'))
            print('_'*70)

            for ii in range(self.SI.shape[0]):
                print(template.format(ii,
                                      np.round(self.SI[ii], 3),
                                      np.round(self.SR[ii], 3),
                                      np.round(self.DW[ii], 3),
                                      np.round(self.AIC[ii], 3),
                                      np.round(self.BIC[ii], 3),
                                      np.round(self.R_square[ii, :].mean(), 3)))
            print('_'*70)
            print(template.format('Total',
                                  np.round(self.SI.mean(), 3),
                                  np.round(self.SR.mean(), 3),
                                  np.round(self.DW.mean(), 3),
                                  np.round(self.AIC.mean(), 3),
                                  np.round(self.BIC.mean(), 3),
                                  np.round(self.R_square.mean(), 3)))
        else:
            template = "{0:<10}{1:<10}{2:<10}{3:<10}{4:<10}{5:<10}"

            print(template.format('SI', 'SR', 'DW', 'AIC', 'BIC', 'R Square'))
            print('_'*60)

            print(template.format(np.round(self.SI, 3),
                                  np.round(self.SR, 3),
                                  np.round(self.DW, 3),
                                  np.round(self.AIC, 3),
                                  np.round(self.BIC, 3),
                                  np.round(self.R_square.mean(), 3)))


__all__.append('ModelDiagnostics')
register_class(ModelDiagnostics)



#!/usr/bin/python

import sys
from os import path
from setuptools import setup

# Scripts
scripts = []

# Local hack
if len(sys.argv) == 2 and sys.argv[1] == '--list-scripts':
    print(' '.join(scripts))
    sys.exit(0)

# read the contents of your README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

name = 'sails'
version = '1.1'
release = '1.1.1'

setup(
    name=name,

    version=release,

    description='Spectral Analysis In Linear Systems',

    # this becomes the PyPI landing page
    long_description=long_description,
    long_description_content_type='text/x-rst',

    # Author details
    author='Andrew Quinn <andrew.quinn@psych.ox.ac.uk>, '
           'Mark Hymers <mark.hymers@ynic.york.ac.uk>',
    author_email='sails-devel@ynic.york.ac.uk',

    # Choose your license
    license='GPL-2+',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
      'Development Status :: 4 - Beta',

      # Indicate who your project is intended for
      'Intended Audience :: Science/Research',
      'Topic :: Scientific/Engineering :: Bio-Informatics',
      'Topic :: Scientific/Engineering :: Information Analysis',
      'Topic :: Scientific/Engineering :: Mathematics',

      # Pick your license as you wish (should match "license" above)
      'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',

      # Specify the Python versions you support here. In particular, ensure
      # that you indicate whether you support Python 2, Python 3 or both.
      'Programming Language :: Python :: 2',
      'Programming Language :: Python :: 2.7',
      'Programming Language :: Python :: 3',
      'Programming Language :: Python :: 3.4',
      'Programming Language :: Python :: 3.5',
    ],

    keywords='multivariate autoregressive models spectral',

    packages=['sails', 'sails.tests'],

    install_requires=['numpy',
                      'scipy',
                      'matplotlib',
                      'h5py',
                      'sphinx_rtd_theme'],

    extras_require={
        'test': ['coverage'],
    },

    command_options={
            'build_sphinx': {
                'project': ('setup.py', name),
                'version': ('setup.py', name),
                'release': ('setup.py', name)}},

    package_data={'sails': ['support/*.csv']},
)



#!/usr/bin/python

# vim: set expandtab ts=4 sw=4:

import math
import numpy as np
from scipy import signal, stats

from .anam import AbstractAnam, register_class

__all__ = []


def fast_resample(X, ds_factor, axis=1):
    """Fast resampling of a timeseries array. This function pads the
    time-dimension to the nearest power of 2 to ensure that the resampling can
    use efficient FFT routines. The padding is removed before the downsampled
    data are returned.

    Parameters
    ----------
    X : ndarray
        Data to be resampled

    ds_factor : float
        Factor of downsampling to apply. Output sample rate will be sample_rate/ds_factor

    axis : int
        Axis along which to perform resampling.  Defaults to 1 for compatibility reasons

    Returns
    -------
    ndarray
        Resampled dataset, of size [nchannels x nsamples]

    """

    orig_size = X.shape[axis]
    target_size = int(X.shape[axis] // ds_factor)

    tmp_size = int(2**math.ceil(math.log(X.shape[axis], 2)))
    tmp_target_size = int(tmp_size // ds_factor)

    # We zero-pad the data in the time dimension up to the temporary size
    # Start by creating an array of appropriate shape
    new_shape = list(X.shape)
    new_shape[axis] = tmp_size

    tmp_X = np.zeros(new_shape, dtype=X.dtype)

    # Now we copy the data into the relevant part of the temporary array
    slic = [slice(None)] * len(X.shape)
    slic[axis] = slice(0, orig_size)
    slic = tuple(slic)

    tmp_X[slic] = X

    # Finally, we resample along the relevant axis
    resampled_X = signal.resample(tmp_X, tmp_target_size, axis=axis)

    # And extract just the parts we want
    slic = [slice(None)] * len(X.shape)
    slic[axis] = slice(0, target_size)
    slic = tuple(slic)

    return resampled_X[slic]


__all__.append('fast_resample')


def epoch_data(X, segment_len=None, trials=None):
    """Epoch a continuous data array. Takes a 2d [channels x samples] array and
    returns a [channels x segment_len x ntrials] array.

    Data can EITHER be epoched by specifying a segment_len value or a trials
    array. If a segment_len is passed then the samples array is split into
    none-overlapping segments of the specified length. Any samples not fitting
    into the final epoch are dropped. If a trials array (of shape [ntrials x 2]
    each row containing a start_ind and a stop_ind) then X is split into the
    trails it specifies.

    Parameters
    ----------
    X : 2d data array
        Data to be epoched
    segment_len : int > 0
        Length of continuous data segments to split dataset into
    trials : 2d array
        Trial start and stop indices contained in an [ntrials x 2] array

    Returns
    -------
    3d array
        Epoched dataset

    """

    if (X.ndim == 3) and (X.shape[0] > 0):
        raise ValueError('Data is already epoched')

    if (segment_len is not None) and (trials is not None):
        raise ValueError("Please specify either 'segment_len' or 'trials'. Can only apply one or the other")

    if segment_len is not None:
        nsegs = X.shape[1] // segment_len
        # Doing the reshape with FORTRAN ordering to get the desired behaviour.
        # Force the array back to be memory-contiguous afterwards. (Might be a better way to do this)
        Xout = X[:, :nsegs * segment_len].reshape(X.shape[0], segment_len, nsegs, order='F')
        return np.ascontiguousarray(Xout)

    if trials is not None:
        if np.any(trials < 0):
            raise ValueError("Negative trial index found, 'trials' should be strictly positive")
        if np.any(trials > X.shape[1]):
            outs = trials[trials > X.shape[1]]
            raise ValueError("Trial indices {0} are out of bounds for axis 1 of X (size={1})".format(outs, X.shape[1]))

        trial_len = np.abs(np.diff(trials[0, :])[0])
        Xout = np.zeros((X.shape[0],  trial_len, trials.shape[0]))
        for ii in range(trials.shape[0]):
            Xout[:, :, ii] = X[:, trials[ii, 0]:trials[ii, 1]]
        return Xout


__all__.append('epoch_data')


def get_valid_samples(data, delay_vect, mode='valid', backwards=False):
    """
    Helper function for excluding or replacing data samples which are not
    preceded by a full delay vect of samples for prediction.

    Parameters
    ----------
    data : ndarray
        Array of data to trim or change, of size [nchannels x nsamples x ntrials]
    delay_vect : ndarray
        Vector of lags included in model
    mode : {'valid','full_nan','full'}
        Options for excluding or replacing residuals which do not have a full
        model prediction ie the third sample of an order 5 model. 'valid'
        removes samples without a full model prediction, 'full_nan' returns
        resids of the same size as the data with nans replacing excluded
        samples and 'full' returns resids keeping non-full model samples in
        place. (Default is 'valid')
    backwards : Boolean
        Flag indicating whether to edit the first or last samples

    Returns
    -------
    ndarray
        data with non-valid samples removed or replaced with nans

    """

    if mode == 'full':
        # just pass through
        return data

    # Don't work in place
    ret = data.copy()

    if mode == 'valid':
        if backwards:
            ret = ret[:, :-len(delay_vect), :]
        else:
            ret = ret[:, len(delay_vect):, :]

    elif mode == 'full_nan':
        if backwards:
            ret[:, -len(delay_vect):, :] = np.nan
        else:
            ret[:, :len(delay_vect), :] = np.nan

    else:
        raise ValueError("Residual return mode is not recognised, please use \
                          'valid', 'full_nan' or 'full'")

    return ret


__all__.append('get_valid_samples')


def gesd(x, alpha=0.05, p_out=.1, outlier_side=0):
    """Detect outliers using Generalized ESD test

     Parameters
     ----------
     x : vector
        Data set containing outliers
     alpha : scalar
        Significance level to detect at (default = 0.05)
     p_out : int
        Maximum number of outliers to detect (default = 10% of data set)
     outlier_side : {-1,0,1}
        Specify sidedness of the test
           - outlier_side = -1 -> outliers are all smaller
           - outlier_side = 0  -> outliers could be small/negative or large/positive (default)
           - outlier_side = 1  -> outliers are all larger

    Returns
    -------
    idx : boolean vector
        Boolean array with TRUE wherever a sample is an outlier
    x2 : vector
        Input array with outliers removed

    References
    ----------
    B. Rosner (1983). Percentage Points for a Generalized ESD Many-Outlier Procedure. Technometrics 25(2), pp. 165-172.
    http://www.jstor.org/stable/1268549?seq=1

    """

    if outlier_side == 0:
        alpha = alpha/2

    if not isinstance(x, np.ndarray):
        x = np.asarray(x)

    n_out = int(np.ceil(len(x)*p_out))

    if np.any(np.isnan(x)):
        # Need to find outliers only in finite x
        y = np.where(np.isnan(x))[0]
        idx1, x2 = gesd(x[np.isfinite(x)], alpha, n_out, outlier_side)

        # idx1 has the indexes of y which were marked as outliers
        # the value of y contains the corresponding indexes of x that are outliers
        idx = np.zeros_like(x).astype(bool)
        idx[y[idx1]] = True

    n = len(x)
    temp = x.copy()
    R = np.zeros((n_out,))
    rm_idx = np.zeros((n_out,), dtype=int)
    lam = np.zeros((n_out,))

    for j in range(0, int(n_out)):
        i = j+1
        if outlier_side == -1:
            rm_idx[j] = np.nanargmin(temp)
            sample = np.nanmin(temp)
            R[j] = np.nanmean(temp) - sample
        elif outlier_side == 0:
            rm_idx[j] = int(np.nanargmax(abs(temp-np.nanmean(temp))))
            R[j] = np.nanmax(abs(temp-np.nanmean(temp)))
        elif outlier_side == 1:
            rm_idx[j] = np.nanargmax(temp)
            sample = np.nanmax(temp)
            R[j] = sample - np.nanmean(temp)

        R[j] = R[j] / np.nanstd(temp)
        temp[int(rm_idx[j])] = np.nan

        p = 1-alpha/(n-i+1)
        t = stats.t.ppf(p, n-i-1)
        lam[j] = ((n-i) * t) / (np.sqrt((n-i-1+t**2)*(n-i+1)))

    # Create a boolean array of outliers
    idx = np.zeros((n,)).astype(bool)
    idx[rm_idx[np.where(R > lam)[0]]] = True

    x2 = x[~idx]

    return idx, x2


def _find_outliers_in_dims(X, axis=-1, metric_func=np.std, gesd_args=None):
    """Find outliers across specified dimensions of an array"""

    if gesd_args is None:
        gesd_args = {}

    if axis == -1:
        axis = np.arange(X.ndim)[axis]

    squashed_axes = tuple(np.setdiff1d(np.arange(X.ndim), axis))
    metric = metric_func(X, axis=squashed_axes)

    rm_ind, _ = gesd(metric, **gesd_args)

    return rm_ind


def _find_outliers_in_segments(X, axis=-1, segment_len=100,
                               metric_func=np.std, gesd_args=None):
    """Create dummy-segments in a dimension of an array and find outliers in it"""

    if gesd_args is None:
        gesd_args = {}

    if axis == -1:
        axis = np.arange(X.ndim)[axis]

    # Prepare to slice data array
    slc = []
    for ii in range(X.ndim):
        if ii == axis:
            slc.append(slice(0, segment_len))
        else:
            slc.append(slice(None))

    # Preallocate some variables
    starts = np.arange(0, X.shape[axis], segment_len)
    metric = np.zeros((len(starts), ))
    bad_inds = np.zeros(X.shape[axis])*np.nan

    # Main loop
    for ii in range(len(starts)):
        if ii == len(starts)-1:
            stop = None
        else:
            stop = starts[ii]+segment_len

        # Update slice on dim of interest
        slc[axis] = slice(starts[ii], stop)
        # Compute metric for current chunk
        metric[ii] = metric_func(X[tuple(slc)])
        # Store which chunk we've used
        bad_inds[slc[axis]] = ii

    # Get bad segments
    rm_ind, _ = gesd(metric, **gesd_args)
    # Convert to int indices
    rm_ind = np.where(rm_ind)[0]
    # Convert to bool in original space of defined axis
    bads = np.isin(bad_inds, rm_ind)
    return bads


def detect_artefacts(X, axis=None, reject_mode='dim', metric_func=np.std,
                     segment_len=100, gesd_args=None, ret_mode='bad_inds'):
    """Detect bad observations or segments in a dataset

    Parameters
    ----------
    X : ndarray
        Array to find artefacts in.
    axis : int
        Index of the axis to detect artefacts in
    reject_mode : {'dim' | 'segments'}
        Flag indicating whether to detect outliers across a dimension (dim;
        default) or whether to split a dim into segments and detect outliers in
        the them (segments)
    metric_func : function
        Function defining metric to detect outliers on. Defaults to np.std but
        can be any function taking an array and returning a single number.
    segement_len : int > 0
        Integer window length of dummy epochs for bad_segment detection
    gesd_args : dict
        Dictionary of arguments to pass to gesd
    ret_mode : {'good_inds','bad_inds','zero_bads','nan_bads'}
        Flag indicating whether to return the indices for good observations,
        indices for bad observations (default), the input data with outliers
        removed (zero_bads) or the input data with outliers replaced with nans
        (nan_bads)

    Returns
    -------
    ndarray
        If ret_mode is 'bad_inds' or 'good_inds', this returns a boolean vector
        of length X.shape[axis] indicating good or bad samples. If ret_mode if
        'zero_bads' or 'nan_bads' this returns an array copy of the input data
        X with bad samples set to zero or np.nan respectively.

    """

    if reject_mode not in ['dim', 'segments']:
        raise ValueError("reject_mode: '{0}' not recognised".format(reject_mode))

    if ret_mode not in ['bad_inds', 'good_inds', 'zero_bads', 'nan_bads']:
        raise ValueError("ret_mode: '{0}' not recognised")

    if axis is None or axis > X.ndim:
        raise ValueError('bad axis')

    if reject_mode == 'dim':
        bad_inds = _find_outliers_in_dims(X, axis=axis, metric_func=metric_func, gesd_args=None)

    elif reject_mode == 'segments':
        bad_inds = _find_outliers_in_segments(X, axis=axis,
                                              segment_len=segment_len,
                                              metric_func=metric_func,
                                              gesd_args=None)

    if ret_mode == 'bad_inds':
        return bad_inds
    elif ret_mode == 'good_inds':
        return bad_inds == False  # noqa: E712
    elif ret_mode in ['zero_bads', 'nan_bads']:
        out = X.copy()

        slc = []
        for ii in range(X.ndim):
            if ii == axis:
                slc.append(bad_inds)
            else:
                slc.append(slice(None))
        slc = tuple(slc)

        if ret_mode == 'zero_bads':
            out[slc] = 0
            return out
        elif ret_mode == 'nan_bads':
            out[slc] = np.nan
            return out


class PCA(AbstractAnam):
    """
    Class for handling PCA-based reduction before fitting a model
    """

    hdf5_outputs = ['npcs', 'data_mean', '_U', '_s', '_VT',
                    'explained_variance', 'explained_variance_ratio',
                    'components', 'loadings', 'scores']

    def __init__(self, data=None, npcs=None):
        AbstractAnam.__init__(self)

        # Need to cope with no-argument initialisation for anamnesis
        if data is not None:
            self.data_mean = data.mean(axis=0)
            self._pca_svd(data, npcs)

    def _pca_svd(self, data, npcs=10):
        """
        Compute a Principal Components Analysis on a given dataset using the SVD
        method.

        Parameters
        ----------
        X : ndarray
            Input data of shape [nsamples x nfeatures]. The second dimension is
            reduced by the PCA.

        Returns
        -------
        scores : ndarray [nsamples x npcs]
            The dimensionality-reduced data. This contains the value of each
            observation in the new PCA-space.

        components : ndarray [npcs x nfeatures]
            The eigenvectors describing how each feature (node or connection) loads
            onto the latent PCA dimensions.

        loadings : ndarray [npcs x nfeatures]
            The components scaled by their contribution to the variance in the
            original data

        explained_variance_ratio : ndarray [npcs]
            The proportion of variance explained by each PC

        """

        self.npcs = npcs

        # Demean observations
        self.data_mean = data.mean(axis=0)[None, :]

        # compute
        self._U, self._s, self._VT = np.linalg.svd(data - self.data_mean, full_matrices=False)

        # Variance explained metrics
        var = self._s ** 2 / (data.shape[0]-1)
        self.explained_variance = var[:npcs]
        self.explained_variance_ratio = var[:npcs] / var.sum()

        # The weights for each original variable in each principal component
        self.components = self._VT[:npcs, :]  # Eigenvectors
        self.loadings = self.components * self._s[:npcs, None]  # Scaled by variance contributed to data

        # The new co-ordinates of each observation in PCA-space
        self.scores = self._s[:npcs] * self._U[:, :npcs]

        # Check that component self.scores are properly decorrelated
        try:
            C = np.corrcoef(self.scores.T)
            assert(np.sum(np.abs(C - np.diag(np.diag(C)))) < 1e-10)
        except AssertionError:
            print('Observations are not properly de-correlated by PCA. Data demeaning might have gone wrong.')
            raise

    def project_score(self, scores):
        """
        Compute a projection of a set of scores back to original data space
        """
        return self.data_mean + np.dot(scores, self.components)


__all__.append('PCA')
register_class(PCA)



#!/usr/bin/python

# vim: set expandtab ts=4 sw=4:

import scipy
from scipy import signal
import numpy as np
import matplotlib.pyplot as plt

from .anam import AbstractAnam, register_class
from sails.support import ensure_leading_negative
from sails.modelfit import coloured_noise_fit

__all__ = []


def adjust_phase(x):
    """Method for normalising eigenvector matrices to meet the conditions in
    eqn 21 of [Neumaier2001]_.

    Parameters
    ----------
    x : ndarray
        Vector of complex numbers to adjust the phase of

    Returns
    -------
    type
        Vector of numbers with adjusted phase

    """

    ox = np.zeros_like(x)

    # Rows of x
    for j in range(x.shape[0]):
        a = np.atleast_2d(np.real(x[j, :])).T
        b = np.atleast_2d(np.imag(x[j, :])).T
        tmp = np.dot(b.T, b) - np.dot(a.T, a)
        phi = .5 * np.arctan(2 * np.sum(a*b / tmp))
        bnorm = np.linalg.norm(np.sin(phi) * a + np.cos(phi) * b)
        anorm = np.linalg.norm(np.cos(phi) * a + np.sin(phi) * b)

        if bnorm > anorm:
            if phi < 0:
                phi = phi - (np.pi/2.)
            else:
                phi = phi + (np.pi/2.)

        ox[j, :] = x[j, :] * np.exp((0+1j) * phi)

    return ox


__all__.append('adjust_phase')


class MvarModalDecomposition(AbstractAnam):
    """Object which calculates the modal decomposition of a fitted model
    derived from the AbstractLinearModel class. This is typically created using
    the :meth:`initialise` classmethod which computes the decomposition and returns a
    populated object instance.
    """

    # This is only used when used within Anamnesis
    hdf5_outputs = ['nsignals', 'order', 'companion', 'evals', 'evecsr',
                    'evecsl', 'pole_pairs', 'modeinds', 'period',
                    'dampening_time', 'peak_frequency', 'dampening_time_hz', 'nmodes',
                    'delay_vect']

    def __init__(self):
        AbstractAnam.__init__(self)

    @classmethod
    def initialise(cls, model, sample_rate=None, normalise=False):
        """Compute the modal pole-residue decomposition of the A matrix from a
        fitted MVAR model.

        Currently only implemented for a single realisation (A.ndim == 3)

        Parameters
        ----------
        model : LinearModel
            Fitted model object.  Must be derived from AbstractLinearModel.
        sample_rate : float
            sample rate of system in Hz (Default value = None)
        normalise : bool
            Whether to adjust the phase of the eigenvectors (Default value = False)

        Returns
        -------
        type
            Modal decomposition object

        """

        # Create object
        ret = cls()

        A = model.parameters
        nsignals = ret.nsignals = A.shape[0]
        ret.order = A.shape[2] - 1
        nmodes = nsignals * ret.order

        if A.ndim == 4:
            pass
        else:
            # Add dummy epoch
            A = A[..., None]

        A = ensure_leading_negative(A)

        if model.resid_cov.ndim == 2:
            # Add a dummy epoch
            model.resid_cov = model.resid_cov[..., None]

        nrealisations = A.shape[3]

        # preallocate arrays
        ret.evals = np.zeros((nmodes, nrealisations), dtype=complex)
        ret.evecsr = np.zeros((nmodes, nmodes, nrealisations), dtype=complex)
        ret.evecsl = np.zeros((nmodes, nmodes, nrealisations), dtype=complex)

        ret.modeinds = np.zeros((nmodes, nrealisations), dtype=int)
        ret.pole_pairs = np.zeros_like(ret.evals, dtype=int) * np.nan

        for ii in range(nrealisations):

            # Generate companion form
            from .modelfit import A_to_companion
            ret.companion = A_to_companion(A[:, :, 1:, ii])

            # Roots are now the eigenvalues of the companion matrix
            ret.evals[:, ii], ret.evecsl[:, :, ii], ret.evecsr[:, :, ii] = \
                scipy.linalg.eig(ret.companion, left=True, right=True)

            # Apply normalisation from http://climate-dynamics.org/wp-content/uploads/2015/08/arfit.pdf
            if normalise:
                ret.evecsr[:, :, ii] = adjust_phase(ret.evecsr[:, :, ii].T).T

            # take left eigenvecs from right, might preserve scaling better
            ret.evecsl[:, :, ii] = np.linalg.inv(ret.evecsr[:, :, ii].conj().T)

            ret.modeinds[:, ii] = np.argsort(np.abs(ret.evals[:, ii]))[::-1]

            # Loop through poles with an imaginary part and find pairings, poles
            # with a zero imaginary part are left indexed as nan
            ppidx = 0
            for idx in range(len(ret.pole_pairs)):
                if ret.evals[idx].imag > 0:
                    # Find the conjugate pair
                    cidx = (abs(ret.evals[idx] - ret.evals.conj())).argmin()
                    ret.pole_pairs[idx] = ppidx
                    ret.pole_pairs[cidx] = ppidx
                    ppidx += 1

        # Compute the period and dampening time per mode
        # We ignore /0 warnings as they are expected for real modes
        old_warning = np.geterr()['divide']
        np.seterr(divide='ignore')
        ret.period = 2*np.pi / np.abs(np.angle(ret.evals))
        ret.dampening_time = -1.0 / np.log(np.abs(ret.evals))
        np.seterr(divide=old_warning)

        if sample_rate is not None:
            ret.peak_frequency = sample_rate / ret.period
            ret.dampening_time_hz = sample_rate / ret.dampening_time
        else:
            ret.peak_frequency = None
            ret.dampening_time_hz = None

        ret.nmodes = nmodes
        ret.delay_vect = model.delay_vect

        return ret

    @property
    def period_hz(self):
        """This is an old and deprecated way of accessing peak_frequency"""
        return self.peak_frequency

    def compute_residues(self):
        """Compute the residue matrix from the eigenvectors associated with each
        pole.

        Returns
        -------
        ndarray
            Array containing residue matrices for each pole.

        """

        # Compute the residue matrices and frequency responses per mode
        residue = np.zeros((self.nsignals, self.nsignals, self.nmodes), dtype=complex)

        for imode in range(len(self.evals)):
            # Compute residue matrix from left and right eigenvectors
            l = np.atleast_2d(self.evecsl[:self.nsignals, imode]).conj()
            r = np.atleast_2d(self.evecsr[:self.nsignals, imode]).T
            residue[:, :, imode] = l.dot(r).T

        return residue

    def compute_modal_parameters(self):
        """Compute the order 1 or 2 autoregressive coefficients associated with
        each pole or conjugate pair.

         .. warning::

            This function is a work in progress and has not been fully validated.

        Returns
        -------
        ndarray
            Array containing the time-domain AR parameters for each mode

        """

        # Compute the residue matrices and frequency responses per mode
        residue = self.compute_residues()

        A = np.zeros((self.nsignals, self.nsignals, 3, self.pole_pairs.max()+1))

        # Loop through nodes and connections
        for ii in np.arange(self.nsignals):
            for jj in np.arange(self.nsignals):
                # For each pole pair
                for kk in np.arange(self.pole_pairs.max()+1):
                    inds = self.pole_pairs[:, 0] == kk
                    b, a = signal.invresz(residue[ii, jj, inds].squeeze(),
                                          self.evals[inds, 0], [0])

                    # I'm scaling the ar coefficients by the leading numerator coeff.
                    # This scaling doesn't affect the roots (ie peak frequency) but
                    # will show up in the relative measures.
                    A[ii, jj, :len(a), kk] = a * b[0]

        return A

    def modal_freqz(self):
        """Compute filter characteristics for each pole"""

        # Compute the frequency response for the whole system
        w, h = signal.freqz(1, self.evals)
        mag_response = np.abs(h)
        phase_response = np.unwrap(np.angle(h))

        return w, mag_response, phase_response

    def excitation(self, resid_cov):
        """Compute the mode excitation as defined in [Neumaier2001]_.
        This is a metric to  quantify the dynamical importance of each mode in
        the decomposition.

        .. warning::

            This function is a work in progress and has not been fully validated.

        Parameters
        ----------
        resid_cov : ndarray
            Residual covariance matrix for modelled system

        Returns
        -------
        ndarray
            Vector containing modal excitation values

        """

        # Compute the excitation per mode
        inv_evecsr = np.linalg.inv(self.evecsr)

        decoupled_cov = inv_evecsr[:, :self.nsignals].dot(resid_cov[:, :, 0])
        decoupled_cov = decoupled_cov.dot(inv_evecsr[:, :self.nsignals].conj().T)

        abs_lambda_sq = np.power(np.abs(self.evals), 2)

        excitation = (np.diag(decoupled_cov) / (1 - abs_lambda_sq)).real
        excitation = excitation / np.sum(excitation)

        return excitation

    def transfer_function(self, sample_rate, freq_vect, modes=None,
                          sum_modes=True):
        """Compute the transfer function in pole-residue form, splitting the
        system into modes with separate transfer functions. The full system
        transfer function is then a linear sum of each modal transfer function.

        When sum_modes is False, this function returns the transfer function
        for each individual pole (ie will return a transfer function for each
        pole in a conjugate pair). Please use per_mode_transfer_function to
        compute the transfer function for individual modes (ie one transfer
        function for each real pole or complex-conjugate pair)

        Parameters
        ----------
        sample_rate : float
            sample rate of system in Hz
        freq_vect : ndarray
            Vector of frequencies at which to evaluate function
        modes : list of ints
            List of mode indices to evaluate over (optional) (Default value = None)
        sum_modes : bool
            Boolean indicating whether to sum modes or return all (Default value = True)

        Returns
        -------
        ndarray
            Array containing the transfer function for each individual pole

        """

        if (sample_rate is None and freq_vect is not None) or \
           (sample_rate is not None and freq_vect is None):
            raise ValueError('Please define both sample_rate and '
                             'freq_vect (or neither to return '
                             'normalised frequency 0->pi)')

        if freq_vect is None and sample_rate is None:
            # Use normalised frequency
            freq_vect_rads = np.linspace(0, np.pi, 512)
        else:
            # convert user defined frequecies to radians
            freq_vect_rads = (freq_vect / (sample_rate / 2.)) * np.pi

        # Allow us to pick out which modes we want
        if modes is None:
            modes = list(range(len(self.evals)))

        if sum_modes is False:
            nmodes = len(modes)
        else:
            nmodes = 1

        # Compute transfer function per mode
        residue = self.compute_residues()
        H = np.zeros((residue.shape[0],
                      residue.shape[0],
                      len(freq_vect_rads),
                      nmodes), dtype=complex)

        # Compute point on unit circle for each frequency
        z = np.cos(freq_vect_rads) + 1j*np.sin(freq_vect_rads)

        idx = 0

        for imode in modes:

            for ifreq in range(len(freq_vect_rads)):

                # Compute transfer function for this frequency and this mode
                num = residue[:, :, imode] * z[ifreq]
                dom = z[ifreq] - self.evals[imode]

                if sum_modes:
                    H[:, :, ifreq, 0] += num / dom
                else:
                    H[:, :, ifreq, idx] = num / dom

            idx += 1

        return H

    def per_mode_transfer_function(self, sample_rate, freq_vect):
        """Extracts the transfer function for each mode by summing the transfer
        function across pole-pairs where necessary

        The transfer function is computed for each pole using
        :meth:`transfer_function` before summing individual
        modes (real valued poles or complex-conjugate pairs).

        Parameters
        ----------
        sample_rate :
            sample rate of system in Hz
        freq_vect :
            Vector of frequencies at which to evaluate function

        Returns
        -------
        ndarray
            Array containing the transfer function for each individual mode

        """

        H = self.transfer_function(sample_rate, freq_vect, sum_modes=False)

        mi = self.mode_indices

        ret = np.zeros((len(freq_vect), len(mi)), dtype=complex)

        for idx, i in enumerate(mi):
            ret[:, idx] = np.sum(H[..., i], axis=3).squeeze()

        return ret

    def modal_transfer_function(self, sample_rate, modes=None):
        """Compute a transfer function matrix based on the excitation period for
        each mode

        Parameters
        ----------
        sample_rate : float
            sample rate of system in Hz
        modes : list of int
            List of mode indices to evaluate over (optional) (Default value = None)

        Returns
        -------
        ndarray
            Transfer function computed for each mode

        """

        # Allow us to pick out which modes we want
        if modes is None:
            modes = list(range(len(self.evals)))

        # Compute transfer function per mode
        H = np.zeros((self.residue.shape[0],
                      self.residue.shape[0],
                      len(modes)), dtype=complex)

        idx = 0

        for imode in modes:
            fv_rad = (self.peak_frequency[imode] / (sample_rate / 2.)) * np.pi

            # Compute point on unit circle for the relevant frequency
            z = np.cos(fv_rad) + 1j*np.sin(fv_rad)

            # Compute transfer function for this frequency and this mode
            num = self.residue[:, :, imode] * z
            dom = z - self.evals[imode]
            H[:, :, idx] = num / dom

            idx += 1

        return H

    def get_mode_inds(self, fmin=0, fmax=None,
                      mag_thresh=0, resid_thresh=0,
                      index_mode='inclusive'):
        """Helper function to identify mode indices based on given criteria.

        Parameters
        ----------
        fmin : float
             Minimum frequency of modes to include (Default value = 0)
        fmax : float
             Maximum frequency of modes to include (Default value = None)
        mag_thresh : float ( 0 > mag_thresh > 1 )
            Minimum pole magnitude to include (Default value = 0)
        resid_thresh : float
            Minimum value  of mode residue norm to include (Default value = 0)
        index_mode : {'inclusive','exclusive'}
            Flag indicating whether mode selection limits should be inclusive
            or exclusive (Default value = 'inclusive')

        Returns
        -------
        ndarray
            Array of integers indexing the included poles

        """

        if fmax is None:
            fmax = self.freq_vect_hz[-1]/2

        if index_mode == 'inclusive':

            inds = (self.peak_frequency >= fmin) * (self.peak_frequency <= fmax)

            if mag_thresh is not None:
                inds = inds * (np.abs(self.evals) >= mag_thresh)

            if resid_thresh is not None:
                inds = inds * (np.abs(self.residue_norm[:, None]) >= resid_thresh)

        elif index_mode == 'exclusive':

            inds = (self.peak_frequency > fmin) * (self.peak_frequency < fmax)

            if mag_thresh is not None:
                inds = inds * (np.abs(self.evals) > mag_thresh)

            if resid_thresh is not None:
                inds = inds * (np.abs(self.residue_norm[:, None]) > resid_thresh)

        else:
            raise ValueError("index_mode {0} not recognised, please use"
                             "'inclusive' or 'exclusive'".format(index_mode))

        return np.where(inds)[0]

    # TODO: Validate this function
    def modal_timecourse(self, data):
        """Compute the modal time-course from a dataset and a fitted model. This
        is the transformation of a set of [nsignals x time] data observations
        into modal coordinates of shape [nmodes x time].

        .. warning::

            This function is a work in progress and has not been fully validated.

        Parameters
        ----------
        data : ndarray
            Input data to convert into modal co-ordinates

        Returns
        -------
        ndarray
            Data transformed into modal time-series

        """

        # Preallocate modal time-course array
        r = np.zeros((self.nmodes, data.shape[1], data.shape[2]))

        for idx in range(self.order, data.shape[1]):

            # Get modal time-course for this sample
            for iep in range(data.shape[2]):

                # Generate delay embedded data for this sample
                deldat = np.zeros((self.nsignals*(self.order-1),))
                for idelay in range(1, len(self.delay_vect)):
                    deldat[(idelay-1)*self.nsignals:idelay*self.nsignals] = \
                        data[:, idx-idelay, iep]
                    r[:, idx, iep] = self.companion.dot(deldat)

        return r

    @property
    def residue_norm(self):
        """The matrix-norm of the residue matrix of each mode"""
        nmodes = len(self.modeinds)
        residue = self.compute_residues()
        return np.linalg.norm(residue.reshape(-1, nmodes), axis=0)

    @property
    def mode_indices(self):
        """Returns a list of tuples of mode indices where each tuple contains the
        indices of either a pole-pair or an unpaired pole

        Returns
        -------
        List
            list of tuples containing indices into the modes

        """
        modes = []
        handled_nodes = np.zeros_like(self.pole_pairs)
        for k in range(self.pole_pairs.shape[0]):
            # Deal with nodes we've already handled as conjugate pairs
            if handled_nodes[k] == 1:
                continue

            # Is this a real mode
            if np.isnan(self.pole_pairs[k]):
                # Single, real mode
                modes.append((k,))
                handled_nodes[k] = 1
            else:
                # Need to find pair index
                pidx = self.pole_pairs[k]

                indices = np.where(self.pole_pairs == pidx)[0]
                modes.append(tuple(indices))
                for idx in indices:
                    handled_nodes[idx] = 1

        return modes

    def pole_plot(self, ax=None, normscale=50, plottype='unitcircle'):
        """Plot the poles of the current system.

        Parameters
        ----------
        ax : matplotlib axes handle
            Optional axis on which to plot (Default value = None)
        normscale : float
            Scaling factor for points on plot (Default value = 50)
        plottype : {'unitcircle','eigenspectrum'}
            Flag indicating whether to plot results on a unit-circle or an
            eigenspectrum (Default value = 'unitcircle')

        Returns
        -------
        matplotlib axes handle
            Reference to axes on which plot was drawn

        """
        if ax is None:
            plt.figure()
            ax = plt.subplot(111)

        norm_factors = self.residue_norm / self.residue_norm.max()

        if plottype == 'unitcircle':

            x = np.cos(np.linspace(0, 2*np.pi, 128))
            y = np.sin(np.linspace(0, 2*np.pi, 128))
            plt.grid(True)

            ax.plot(x, y, 'k')

            ax.plot(.75*x, .75*y, 'k--', linewidth=.2)
            ax.plot(.5*x, .5*y, 'k--', linewidth=.2)
            ax.plot(.25*x, .25*y, 'k--', linewidth=.2)

            ax.plot(1.15*x[8:25], 1.15*y[8:25], 'k')
            ax.arrow(1.15*x[23], 1.15*y[23], x[24]-x[23], y[24]-y[23],
                     head_width=.05, color='k')

            handled_nodes = np.zeros_like(self.pole_pairs)

            for k in range(self.pole_pairs.shape[0]):
                # Deal with nodes we've already handled as conjugate pairs
                if handled_nodes[k] == 1:
                    continue

                # Need to find pair index
                pidx = self.pole_pairs[k]

                # Deal with real pole case
                if np.isnan(pidx):
                    indices = [k]
                else:
                    indices = np.where(self.pole_pairs == pidx)[0]

                # Stash the real and imaginary parts
                r = []
                i = []
                ms = None

                for idx in indices:
                    handled_nodes[idx] = 1

                    r.append(self.evals[idx].real)
                    i.append(self.evals[idx].imag)

                    # The scaling should be the same for the conjugate pair
                    if ms is None:
                        ms = norm_factors[idx]*normscale

                # Plot them all together to maintain the same colour
                # for each pair
                plt.plot(r, i, '+', ms=ms, markeredgewidth=2)

            ax.set_xlabel('Real')
            ax.set_ylabel('Imaginary')

            ax.set_ylim(-1.2, 1.2)
            ax.set_xlim(-1.2, 1.2)

            ax.annotate('Frequency', xy=(.56, 1.02))

            ax.set_aspect('equal')

        elif plottype == 'eigenspectrum':

            mags = np.abs(self.evals)
            mags = mags / (1-mags)

            if self.peak_frequency is None:
                x_vect = self.period
                xlabel = 'Period (samples)'
            else:
                x_vect = self.peak_frequency
                xlabel = 'Frequency (Hz)'

            yt = np.linspace(0, 1., 11)
            yt_scaled = yt / (1-yt)

            ax.scatter(x_vect, mags, c='k', marker='o', s=norm_factors*normscale)
            ax.set_yticks(yt_scaled[:-2])
            ax.set_yticklabels(yt[:-2])
            ax.grid(True)
            ax.set_xlabel(xlabel)
            ax.set_ylabel('Pole Magnitude')

        return ax


__all__.append('MvarModalDecomposition')
register_class(MvarModalDecomposition)


def find_noise_poles(modes, model, X, sample_rate, metric='diff', include_real_poles=True):
    """Helper function for identifying noise poles using a coloured noise fit
    for reference.

    .. warning::

        This is a work in progress and has not been fully validated.

    Parameters
    ----------
    modes : sails MvarModalDecomposition
        Object containing modal decomposition of system
    model : sails LinearModel
        Object containing fitted autoregressive model
    X : ndarray
        Data on to compute the threshold for
    sample_rate : float
        The sampling frequency of the system
    metric : {'diff','diff_noisenorm','diff_datanorm','diff_sumnorm'}
        Method for comparing coloured noise poles to unconstrained poles
        (Default value = 'diff')
    include_real_poles : bool
        Flag indicating whether to include real valued poles (Default value = True)

    Returns
    -------
    tuple
        tuple containing:
        indices to included modes
        indices to excluded noise modes
        metric value for each mode

    """

    coloured_noise, f_alphas = coloured_noise_fit(X, model.order)
    coloured_modes = MvarModalDecomposition.initialise(coloured_noise, sample_rate=sample_rate)

    from scipy import interpolate
    hz, I = np.unique(coloured_modes.peak_frequency[:, 0], return_index=True)
    pchip = interpolate.pchip(coloured_modes.peak_frequency[I, 0],
                              np.abs(coloured_modes.evals[I, 0]))

    I = np.argsort(modes.peak_frequency[:, 0])
    interp_poles = pchip(modes.peak_frequency[:, 0])
    orig_poles = np.abs(modes.evals[:, 0])

    if metric == 'diff':
        vals = (orig_poles-interp_poles)
    elif metric == 'diff_noisenorm':
        vals = (orig_poles-interp_poles) / interp_poles
    elif metric == 'diff_datanorm':
        vals = (orig_poles-interp_poles) / orig_poles
    elif metric == 'diff_sumnorm':
        vals = (orig_poles-interp_poles) / (orig_poles+interp_poles)

    thresh = 0
    if include_real_poles:
        good_modes = np.where(vals > thresh)[0]
    else:
        good_modes = np.where((vals > thresh) & (np.abs(modes.evals[:, 0].imag) > 1e-10))[0]
    bad_modes = np.setdiff1d(np.arange(modes.nmodes), good_modes)

    return good_modes, bad_modes, vals


def get_noise_thresh_gamma(modes, fmin, fmax):
    """Helper function for identifying noise modes by fitting a gamma
    distribution the pole distributions within frequency bins.

    .. warning::

        This is a work in progress and has not been fully validated.

    Parameters
    ----------
    modes : sails MvarModalDecomposition
        Object containing computed modal decomposition
    fmin : float
        Minimum frequency to include in thresholding
    fmax : float
        Maximum frequency to include in thresholding

    Returns
    -------
    tuple
        tuple containing
        centre frequency of bins
        computed threshold for each bin
        interpolated threshold for each pole

    """

    from scipy.stats import gamma

    x = np.linspace(0, 1, 1024)
    freq_centres = np.linspace(fmin, fmax, 64)
    thresh = np.zeros_like(freq_centres)

    for ii in range(len(freq_centres)):
        inds = modes.get_mode_inds(fmin=freq_centres[ii]-2,
                                   fmax=freq_centres[ii]+2)
        m = np.abs(modes.evals[inds])
        mle_tuple = gamma.fit(1-m, 1)
        rv1 = gamma.pdf(x=x, a=mle_tuple[0], loc=mle_tuple[1], scale=mle_tuple[2])
        thresh[ii] = 1-x[np.argmax(rv1)]

    from scipy import interpolate
    f = interpolate.interp1d(freq_centres, thresh, bounds_error=False)

    return freq_centres, thresh, f(modes.peak_frequency)


def cluster_modes(X, n_clusters, pca_dims):
    """Helper function for computing k-means clustering on modes.

    .. warning::

        This is a work in progress and has not been fully validated.

    Parameters
    ----------
    X : ndarray
        Data to cluster, typically modal eigenvectors
    n_clusters : int
        Number of clusters to compute
    pca_dims : int
        Number of PCA components to reduce data to

    Returns
    -------
    ndarray
        Categorical cluster labels
    ndarray
        Inertia for each point

    """

    from sklearn.decomposition import PCA
    from sklearn.cluster import KMeans

    pc = PCA(n_components=pca_dims, whiten=True, copy=True).fit(X)
    km = KMeans(n_clusters=n_clusters)
    clusters = km.fit_predict(pc.components_.T)

    return clusters, km.inertia_


def get_cluster_average(clusters, vals):
    """Helper function for computing an average across defined clusters

    .. warning::

        This is a work in progress and has not been fully validated.

    Parameters
    ----------
    clusters : ndarray
        Categorical cluster labels
    vals : ndarray
        Value to average within clusters

    Returns
    -------
    ndarray
        Average values within clusters

    """

    out = np.zeros((vals.shape[:-1], clusters.max()+1))
    for c in range(clusters.max()+1):
        out[..., c] = vals[..., clusters == c].mean(axis=-1)

    return out
